<?php
/**
 * eACL stream wrapper: shows eacl as directory structure with symlinks to the original dirs
 *
 * @link http://www.egroupware.org
 * @license http://opensource.org/licenses/gpl-license.php GPL - GNU General Public License
 * @package epl
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-16 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

use EGroupware\Api;
use EGroupware\Api\Vfs;

/**
 * eACL stream wrapper: shows eacl as directory structure with symlinks to the original dirs
 *
 * The first level directorires are groups the user is a member off or to which subdirs
 * he has access, which have eACL grants from other user/groups.
 * The second level are symlinks to that directories.
 *
 * The wrapper understands the following query-parameter / mount-options:
 * - ignore=/path/to/ignore (multiple pathes possible: ignore[]=/path1&ignore[]=/path2)
 *
 * Mount example:
 *
 * filemanager/cli.php mount --user root_xxx --password xxx \
 * 		'stylite.eacl://default/?ignore[]=/Verwaltung&ignore[]=/Admins' /Einrichtungen
 *
 * This will ignore (avoid to display) the eACL grants for groups Admins and Verwaltung,
 * but not the ones from the direct subdirs of this grants!
 *
 * The interface is according to the docu on php.net
 *
 * @link http://www.php.net/manual/en/function.stream-wrapper-register.php
 */
class stylite_eacl_stream_wrapper implements Vfs\StreamWrapperIface
{
	/**
	 * Scheme / protocol used for this stream-wrapper
	 */
	const SCHEME = 'stylite.eacl';
	/**
	 * How much should be logged to the apache error-log
	 *
	 * 0 = Nothing
	 * 1 = only errors
	 * 2 = all function calls and errors (contains passwords too!)
	 */
	const LOG_LEVEL = 1;
	/**
	 * optional context param when opening the stream, null if no context passed
	 *
	 * @var mixed
	 */
	//var $context;

	/**
	 * directories and symlinks of this wrapper
	 *
	 * @var array
	 */
	static $fs = array();
	/**
	 * Contenct of the opended directory
	 *
	 * @var array
	 */
	protected $opened_dir;

	/**
	 * This method is called immediately after your stream object is created.
	 *
	 * @param string $path URL that was passed to fopen() and that this object is expected to retrieve
	 * @param string $mode mode used to open the file, as detailed for fopen()
	 * @param int $options additional flags set by the streams API (or'ed together):
	 * - STREAM_USE_PATH      If path is relative, search for the resource using the include_path.
	 * - STREAM_REPORT_ERRORS If this flag is set, you are responsible for raising errors using trigger_error() during opening of the stream.
	 *                        If this flag is not set, you should not raise any errors.
	 * @param string $opened_path full path of the file/resource, if the open was successfull and STREAM_USE_PATH was set
	 * @return boolean true if the ressource was opened successful, otherwise false
	 */
	function stream_open ( $path, $mode, $options, &$opened_path )
	{
		unset($path, $mode, $options, $opened_path);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called when the stream is closed, using fclose().
	 *
	 * You must release any resources that were locked or allocated by the stream.
	 */
	function stream_close ( )
	{
		return false;
	}

	/**
	 * This method is called in response to fread() and fgets() calls on the stream.
	 *
	 * You must return up-to count bytes of data from the current read/write position as a string.
	 * If there are less than count bytes available, return as many as are available.
	 * If no more data is available, return either FALSE or an empty string.
	 * You must also update the read/write position of the stream by the number of bytes that were successfully read.
	 *
	 * @param int $count
	 * @return string/false up to count bytes read or false on EOF
	 */
	function stream_read ( $count )
	{
		unset($count);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called in response to fwrite() calls on the stream.
	 *
	 * You should store data into the underlying storage used by your stream.
	 * If there is not enough room, try to store as many bytes as possible.
	 * You should return the number of bytes that were successfully stored in the stream, or 0 if none could be stored.
	 * You must also update the read/write position of the stream by the number of bytes that were successfully written.
	 *
	 * @param string $data
	 * @return integer
	 */
	function stream_write ( $data )
	{
		unset($data);	// not used, but required by function signature

		return false;
	}

 	/**
 	 * This method is called in response to feof() calls on the stream.
 	 *
 	 * Important: PHP 5.0 introduced a bug that wasn't fixed until 5.1: the return value has to be the oposite!
 	 *
 	 * if(version_compare(PHP_VERSION,'5.0','>=') && version_compare(PHP_VERSION,'5.1','<'))
  	 * {
 	 * 		$eof = !$eof;
 	 * }
  	 *
 	 * @return boolean true if the read/write position is at the end of the stream and no more data availible, false otherwise
 	 */
	function stream_eof ( )
	{
		return false;
	}

	/**
	 * This method is called in response to ftell() calls on the stream.
	 *
	 * @return integer current read/write position of the stream
	 */
 	function stream_tell ( )
 	{
 		return false;
 	}

 	/**
 	 * This method is called in response to fseek() calls on the stream.
 	 *
 	 * You should update the read/write position of the stream according to offset and whence.
 	 * See fseek() for more information about these parameters.
 	 *
 	 * @param integer $offset
 	 * @param integer $whence	SEEK_SET - Set position equal to offset bytes
 	 * 							SEEK_CUR - Set position to current location plus offset.
 	 * 							SEEK_END - Set position to end-of-file plus offset. (To move to a position before the end-of-file, you need to pass a negative value in offset.)
 	 * @return boolean TRUE if the position was updated, FALSE otherwise.
 	 */
	function stream_seek ( $offset, $whence )
	{
		unset($offset, $whence);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called in response to fflush() calls on the stream.
	 *
	 * If you have cached data in your stream but not yet stored it into the underlying storage, you should do so now.
	 *
	 * @return booelan TRUE if the cached data was successfully stored (or if there was no data to store), or FALSE if the data could not be stored.
	 */
	function stream_flush ( )
	{
		return false;
	}

	/**
	 * This method is called in response to fstat() calls on the stream.
	 *
	 * If you plan to use your wrapper in a require_once you need to define stream_stat().
	 * If you plan to allow any other tests like is_file()/is_dir(), you have to define url_stat().
	 * stream_stat() must define the size of the file, or it will never be included.
	 * url_stat() must define mode, or is_file()/is_dir()/is_executable(), and any of those functions affected by clearstatcache() simply won't work.
	 * It's not documented, but directories must be a mode like 040777 (octal), and files a mode like 0100666.
	 * If you wish the file to be executable, use 7s instead of 6s.
	 * The last 3 digits are exactly the same thing as what you pass to chmod.
	 * 040000 defines a directory, and 0100000 defines a file.
	 *
	 * @return array containing the same values as appropriate for the stream.
	 */
	function stream_stat ( )
	{
		return false;
	}

	/**
	 * This method is called in response to unlink() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to delete the item specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support unlinking!
	 *
	 * @param string $path
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function unlink ( $path )
	{
		unset($path);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called in response to rename() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to rename the item specified by path_from to the specification given by path_to.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support renaming.
	 *
	 * The regular filesystem stream-wrapper returns an error, if $url_from and $url_to are not either both files or both dirs!
	 *
	 * @param string $path_from
	 * @param string $path_to
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function rename ( $path_from, $path_to )
	{
		unset($path_from, $path_to);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called in response to mkdir() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to create the directory specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support creating directories.
	 *
	 * @param string $path
	 * @param int $mode
	 * @param int $options Posible values include STREAM_REPORT_ERRORS and STREAM_MKDIR_RECURSIVE
	 * @return boolean TRUE on success or FALSE on failure
	 */
	static function mkdir ( $path, $mode, $options )
	{
		unset($path, $mode, $options);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called in response to rmdir() calls on URL paths associated with the wrapper.
	 *
	 * It should attempt to remove the directory specified by path.
	 * In order for the appropriate error message to be returned, do not define this method if your wrapper does not support removing directories.
	 *
	 * @param string $path
	 * @param int $options Possible values include STREAM_REPORT_ERRORS.
	 * @return boolean TRUE on success or FALSE on failure.
	 */
	static function rmdir ( $path, $options )
	{
		unset($path, $options);	// not used, but required by function signature

		return false;
	}

	/**
	 * This method is called immediately when your stream object is created for examining directory contents with opendir().
	 *
	 * @param string $path URL that was passed to opendir() and that this object is expected to explore.
	 * @return booelan
	 */
	function dir_opendir ( $path, $options )
	{
		if (!is_array($pos =& self::url2pos($path)))
		{
			$this->opened_dir = null;
			if (self::LOG_LEVEL > 0) error_log(__METHOD__."($path,$options) returns false!");
			return false;
		}
		$this->opened_dir = array_keys($pos);
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($path,$options) ".array2string($this->opened_dir));
		reset($this->opened_dir);

		return true;
	}

	/**
	 * This method is called in response to stat() calls on the URL paths associated with the wrapper.
	 *
	 * It should return as many elements in common with the system function as possible.
	 * Unknown or unavailable values should be set to a rational value (usually 0).
	 *
	 * If you plan to use your wrapper in a require_once you need to define stream_stat().
	 * If you plan to allow any other tests like is_file()/is_dir(), you have to define url_stat().
	 * stream_stat() must define the size of the file, or it will never be included.
	 * url_stat() must define mode, or is_file()/is_dir()/is_executable(), and any of those functions affected by clearstatcache() simply won't work.
	 * It's not documented, but directories must be a mode like 040777 (octal), and files a mode like 0100666.
	 * If you wish the file to be executable, use 7s instead of 6s.
	 * The last 3 digits are exactly the same thing as what you pass to chmod.
	 * 040000 defines a directory, and 0100000 defines a file.
	 *
	 * @param string $url
	 * @param int $flags holds additional flags set by the streams API. It can hold one or more of the following values OR'd together:
	 * - STREAM_URL_STAT_LINK	For resources with the ability to link to other resource (such as an HTTP Location: forward,
	 *                          or a filesystem symlink). This flag specified that only information about the link itself should be returned,
	 *                          not the resource pointed to by the link.
	 *                          This flag is set in response to calls to lstat(), is_link(), or filetype().
	 * - STREAM_URL_STAT_QUIET	If this flag is set, your wrapper should not raise any errors. If this flag is not set,
	 *                          you are responsible for reporting errors using the trigger_error() function during stating of the path.
	 *                          stat triggers it's own warning anyway, so it makes no sense to trigger one by our stream-wrapper!
	 * @return array|boolean array with stat info or false if not found
	 */
	static function url_stat ( $url, $flags )
	{
		$path = parse_url($url,PHP_URL_PATH);

		$pos =& self::url2pos($path);
		if ($pos === false)
		{
			if (self::LOG_LEVEL > 0) error_log(__METHOD__."($url,$flags) returns false!");
			return false;	// not found
		}
		if (is_array($pos))	// we stat a directory
		{
			$stat = array(
				'ino'   => 0,
				'name'  => basename($path),
				'mode'  => Vfs\Sqlfs\StreamWrapper::MODE_DIR | 0555,
				'size'  => 0,
				'uid'   => 0,
				'gid'   => 0,
				'mtime' => time(),
				'ctime' => time(),
				'nlink' => 2,
				// eGW addition to return the mime type
				'mime'  => Vfs::DIR_MIME_TYPE,
			);
		}
		else
		{
			$stat = array(
				'ino'   => 0,
				'name'  => basename($path),
				'mode'  => Vfs::MODE_LINK | 0444,
				'size'  => bytes($pos),
				'uid'   => 0,
				'gid'   => 0,
				'mtime' => time(),
				'ctime' => time(),
				'nlink' => 1,
				// eGW addition to return the mime type
				'mime'  => Vfs\Sqlfs\StreamWrapper::SYMLINK_MIME_TYPE,
			);
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url,$flags) = ".array2string($stat));
		return $stat;
	}

	/**
	 * Get a reference to the static $fs var at a given path
	 *
	 * @param string $url path or url
	 * @param boolean $remove =false should we remove the entry, instead returning it
	 * @return string|array|boolean string with symlink, array of directory or false if not found
	 */
	static protected function &url2pos($url,$remove=false)
	{
		static $ignore_done=null;
		if (!$ignore_done && ($query = strpos($url,'?') !== false ? parse_url($url,PHP_URL_QUERY) : null))
		{
			parse_str($query,$query);
			if (isset($query['ignore']))
			{
				foreach((array)$query['ignore'] as $path)
				{
					self::url2pos($path,true);
				}
				$ignore_done = true;
			}
		}

		$path = $url[0] == '/' ? $url : parse_url($url,PHP_URL_PATH);
		if ($path != '/' && substr($path,-1) == '/') $path = substr($path,0,-1);	// remove trailing slash, eg. from webdav

		if ($path[0] != '/')
		{
			if (self::LOG_LEVEL > 0) error_log(__METHOD__."($url) path is not absolute! ".function_backtrace());
			return false;	// we allow only absolut pathes
		}

		$pos =& self::$fs;
		if ($path != '/')
		{
			$parts = explode('/',substr($path,1));
			foreach($parts as $n => $name)
			{
				if (!is_array($pos) || !isset($pos[$name]))
				{
					if (self::LOG_LEVEL > 0) error_log(__METHOD__."($url) $n-th component '$name' not found! ".function_backtrace());
					return false;	// path not found!
				}
				if ($remove && $n == count($parts)-1)
				{
					unset($pos[$name]);
					return true;
				}
				$pos =& $pos[$name];
			}
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."($url) = ".array2string($pos));

		return $pos;
	}

	/**
	 * This method is called in response to readdir().
	 *
	 * It should return a string representing the next filename in the location opened by dir_opendir().
	 *
	 * @return string
	 */
	function dir_readdir ( )
	{
		if (!is_array($this->opened_dir))
		{
 			if (self::LOG_LEVEL > 0) error_log(__METHOD__."( ) returning false!");
			return false;
		}
		$file = current($this->opened_dir); next($this->opened_dir);
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."( ) = '$file'");

		return $file;
	}

	/**
	 * This method is called in response to rewinddir().
	 *
	 * It should reset the output generated by dir_readdir(). i.e.:
	 * The next call to dir_readdir() should return the first entry in the location returned by dir_opendir().
	 *
	 * @return boolean
	 */
	function dir_rewinddir ( )
	{
		if (!is_array($this->opened_dir)) return false;
		reset($this->opened_dir);
		return true;
	}

	/**
	 * This method is called in response to closedir().
	 *
	 * You should release any resources which were locked or allocated during the opening and use of the directory stream.
	 *
	 * @return boolean
	 */
	function dir_closedir ( )
	{
		$this->opened_dir = null;
		return true;
	}

	/**
	 * This method is called in response to readlink().
	 *
	 * @param string $url
	 * @return string|boolean content of the symlink or false if $url is no symlink (or not found)
	 */
	static function readlink($url)
	{
		if (($pos =& self::url2pos($url)) === false || is_array($pos))
		{
			if (self::LOG_LEVEL > 0) error_log(__METHOD__."('$url') returning false (not found or no link)");

			return false;
		}
		if (self::LOG_LEVEL > 1) error_log(__METHOD__."('$url') = '$pos'");

		return $pos;
	}

	/**
	 * Check if the rights of the direct(!) subdirs of a given base are already added to our fs
	 *
	 * @param string $base path whos eAcl rights should be added too, because it's eacl rights get inhirited
	 */
	protected static function check_add_base($base)
	{
		static $bases = array();	// already processed bases

		if (!$base || in_array($base,$bases) || !($eacls = Vfs::get_eacl($base)))
		{
			return;	// nothing to do
		}
		//echo "<p>".__METHOD__."('$base')</p>\n"; _debug_array($eacls);
		foreach ($eacls as $eacl)
		{
			if (!($account_lid = $GLOBALS['egw']->accounts->id2name($eacl['owner'])))
			{
				continue;	// they already exist --> no need to add them
			}
			self::$fs[$account_lid][Vfs::basename($base)] = $base;

			// include for subs of the root (eg. /Akten) all their children
			if (count(explode('/',$base)) == 2 && ($d = Vfs::opendir($base)))
			{
				while(($file = readdir($d)) && ($sub_eacls = Vfs::get_eacl($path = $base.'/'.$file)))
				{
					//echo "<p>setting $path</p>\n"; _debug_array($sub_eacl);
					foreach($sub_eacls as $sub_eacl)
					{
						if ($sub_eacl == $eacl ||
							!($account_lid = $GLOBALS['egw']->accounts->id2name($sub_eacl['owner'])))
						{
							continue;       // they already exist --> no need to add them
						}
						self::$fs[$account_lid][$file] = $path;
					}
				}
			}
		}
		$bases[] = $base;
	}

	/**
	 * Initializse our static vars
	 *
	 */
	public static function init_static( )
	{
		//$start = microtime(true);

		if (!(self::$fs = Api\Session::getSession(__CLASS__, 'fs')))
		{
			self::$fs = array();
			$memberships = (array) $GLOBALS['egw']->accounts->memberships($GLOBALS['egw_info']['user']['account_id']);
			//_debug_array($memberships);
			foreach($memberships as $account_id => $account_lid)
			{
				if (($eacl = $GLOBALS['egw']->acl->get_all_location_rights($account_id,Vfs\Sqlfs\StreamWrapper::EACL_APPNAME,false)))
				{
					//echo $account_lid; _debug_array($eacl);
					foreach($eacl as $fs_id => $rights)
					{
						if (($path = Vfs\Sqlfs\StreamWrapper::id2path($fs_id)) && $rights)
						{
							self::$fs[$account_lid][Vfs::basename($path)] = $path;
							// check if there's a further inheritance of that path (only direct subdirs)
							self::check_add_base($path);
						}
					}
				}
			}
			Api\Session::setSession(__CLASS__, 'fs', self::$fs);
		}
		//echo "<p>".__METHOD__."() took ".round(1000*(microtime(true)-$start),1)." ms</p>\n";
		//_debug_array(self::$fs);
		stream_register_wrapper(self::SCHEME ,__CLASS__);
	}
}

stylite_eacl_stream_wrapper::init_static();
