<?php
/**
 * EPL Functions - Filemanager notifications
 *
 * @link www.egroupware.org
 * @author rb(at)egroupware.org
 * @copyright (c) 2011-16 by rb(at)egroupware.org
 * @package epl
 * @subpackage filemanager
 * @version $Id$
 */

use EGroupware\Api;
use EGroupware\Api\Vfs;

/**
 * Subscribtions/notifications for filemanager
 *
 * Translations:
 * - Filemanager UI parts must have "filemanager" as app
 * - Notification stuff need to be common, as it can be triggered from all apps
 */
class stylite_filemanager_notifications
{
	/**
	 * Enable / disable debug messages to error_log
	 */
	const DEBUG = false;

	/**
	 * Hooks called by vfs, implemented to be able to notify subscribed users about changed files
	 *
	 * No need to care for rename or unlink/rmdir as subscriptions are store as properties of the file/directory!
	 *
	 * @param array $data
	 * @param string $data[location] 'vfs_read', 'vfs_added', 'vfs_modified', 'vfs_unlink', 'vfs_rename', 'vfs_mkdir', 'vfs_rmdir'
	 * @param string $data[path] vfs path
	 * @param string $data[url]  backend url
	 * @param string $data[mode] mode of fopen for location=vfs_file_modified
	 * @param string $data[to|from|to_url|from_url] for location=vfs_file_rename
	 * @param string $data[stat] only for vfs_(unlink|rmdir), as hook get's called after successful unlink/rmdir call
	 */
	public static function vfs_hooks(array $data)
	{
		if (self::DEBUG) error_log(__METHOD__.'('.array2string($data).')');

		$path = $data['path'];

		// ignore / do NOT notify about temporary files or lockfiles created by office programms
		if (preg_match('/(^~\$|^\.~lock\.|\.tmp$)/',Vfs::basename($path))) return;

		$stat = isset($data['stat']) ? $data['stat'] : Vfs::stat($path);

		// we ignore notifications about zero size files, created by some WebDAV clients prior to every real update!
		if ($stat['size'] && ($subscriptions = self::get_subscriptions($path)) && class_exists('notifications'))
		{
			//error_log(__METHOD__.'('.array2string($data).') subscriptions='.array2string($subscriptions));
			$notify = $email_notify = array();
			// flatten array (path => subscriber pairs --> array of subscribers)
			$subscriptions = call_user_func_array('array_merge', $subscriptions);
			// get ride of multiple subscriptions, resolve groups and check if notified users have access to the file
			foreach($subscriptions as $who)
			{
				if (strpos($who, ':')) list($owner, $who) = explode(':', $who);

				// ignore not longer existing accounts
				if (!$GLOBALS['egw']->accounts->exists(is_numeric($who) ? $who : $owner)) continue;

				if (is_numeric($who) && ($type=$GLOBALS['egw']->accounts->get_type($who)))
				{
					if ($type == 'u')
					{
						if (Vfs::check_access($path, Vfs::READABLE, null, $who)) $notify[$who] = $who;
					}
					else
					{
						foreach($GLOBALS['egw']->accounts->members($who, true) as $member)
						{
							if (Vfs::check_access($path, Vfs::READABLE, null, $member)) $notify[$member] = $member;
						}
					}
				}
				elseif(strpos($who, '@') !== false)
				{
					if (Vfs::check_access($path, Vfs::READABLE, null, $owner)) $email_notify[$who] = $who;
				}
			}
			if ($notify || $email_notify)
			{
				//error_log(__METHOD__.'('.array2string($data).') notifying '.array2string(array_values(array_merge($notify,$email_notify))));
				include_once EGW_INCLUDE_ROOT.'/stylite/inc/class.filemanager_tracking.inc.php';
				$tracking = new filemanager_tracking();
				$stat['path'] = $path;
				if (!isset($stat['mime'])) $stat['mime'] = Vfs::mime_content_type($path);
				if (($props = Vfs::propfind($data['path'],Vfs::DEFAULT_PROP_NAMESPACE)))
				{
					foreach($props as $prop)
					{
						$data[$prop['name']] = $prop['val'];
					}
				}
				if (!isset($data['comment'])) $data['comment'] = null;	// mark it queried, so filemanager_tracking does not query it for each sender
				$stat['subscribers'] = $notify;
				$stat['email_subs'] = $email_notify;
				$stat['event'] = $data['location'];

				switch($data['location'])
				{
					case 'vfs_added':
					case 'vfs_modified':
					case 'vfs_unlink':
						// Set up tracking so signature merge works
						$tracking->id_field = 'path';
						if (!$tracking->do_notifications($stat, null, $data['location'] == 'vfs_unlink'))
						{
							if (self::DEBUG) error_log(__METHOD__."() error calling tracking->do_notifications(".array2string($stat).', null) '.array2string($tracking->errors));
						}
						break;

					case 'vfs_rename':
					case 'vfs_rmdir':
					case 'vfs_mkdir':
					default:
						// simply ignoring them for now, they are not necessary to maintain subscriptions stored as properties
						return;
				}
			}
		}
	}

	/**
	 * Hook to add extra tabs to filemanager_ui::file()
	 *
	 * @param array $data values for keys 'location', 'path', 'content'
	 * @return array|boolean of array with values for keys 'data','etemplate','name','label','help' or false if not supported by backend
	 */
	public static function vfs_extra(array $data)
	{
		if (!isset($GLOBALS['egw_info']['user']['apps']['notifications']))
		{
			return array();	// do not offer subscribtions, if user has not rights for notification app
		}
		$content = $readonlys = $preserve = array();
		$path = $data['content']['path'];

		if ($data['content']['subscribe']['now'])
		{
			//_debug_array($data['content']['subscribe']);
			$who = $data['content']['subscribe']['who'] ? $data['content']['subscribe']['who'] : $data['content']['subscribe']['email'];

			if (self::set_subscription($path, $who))
			{
				$content['msg'] = lang('Subscription added.');
			}
			else
			{
				$content['msg'] = lang('Permission denied!');
			}
			$content['subscribe'] = $preserve['subscribe'] = array();
		}
		if ($data['content']['subscriptions']['delete'])
		{
			list($id) = each($data['content']['subscriptions']['delete']);
			list($owner, $who, $sub_path2) = explode(':', $id, 3);
			$sub_path = base64_decode($sub_path2);
			if (self::delete_subscription($sub_path, $owner.($who?':'.$who:'')))
			{
				$content['msg'] = lang('Subscription deleted.');
			}
			else
			{
				$content['msg'] = lang('Permission denied!');
			}
			$content['subscriptions'] = $preserve['subscriptions'] = array();
		}
		if (($subsciptions = self::get_subscriptions($path)))
		{
			//_debug_array($subsciptions);
			$content['subscriptions'] = $preserve['subscriptions'] = array();
			$n = 1;
			$memberships = $GLOBALS['egw']->accounts->memberships(Vfs::$user,true);
			foreach($subsciptions as $path => $subscribers)
			{
				foreach($subscribers as $owner)
				{
					$who = null;
					if (!is_numeric($who)) list($owner,$who) = explode(':', $owner);

					// show all subscriptions only to root or admins
					// regular users only see own subscriptions or one for groups they are a member of
					if (Vfs::$is_root || Vfs::$is_admin || $owner == Vfs::$user || is_numeric($who) && in_array($who,$memberships))
					{
						$content['subscriptions'][$n++] = array(
							'owner' => $owner,
							'who'   => $who,
							'extra' => $who && is_numeric($who) ? Api\Accounts::username($who) : $who,
							'path'  => $path,
							'path_token' => ($path_b64=base64_encode($path)),
						);
						if (!self::allow_subscribe($owner.($who?':'.$who:'')))
						{
							$readonlys["delete[$owner:$who:$path_b64]"] = true;
						}
					}
				}
			}
		}
		return array(
			array(	// subscribtion tab
				'name' => 'stylite.filemanager.file.subscribe',
				'label' => lang('Subscriptions'),
				'help' => lang('Subscribe to notifications about changed files'),
				'data' => $content,
				'preserve' => $preserve,
				'readonlys' => $readonlys,
			),
		);
	}

	// changing NS requires update script, or we loose all subscriptions!
	const SUBSCRIPTION_NS = 'http://www.stylite.de/';
	const SUBSCRIPTIONS_ATTR = 'subscriptions';

	/**
	 * Get all subscriptions for a given path
	 *
	 * @param string $path
	 * @param boolean $parents_too =true also return parent-dirs
	 * @return array with path => array of account_id[:email] pairs, or just array of account_id[:email], if !$parents_too
	 */
	public static function get_subscriptions($path, $parents_too=true)
	{
		$pathes = array($p=$path);
		if ($parents_too)
		{
			while (($p = Vfs::dirname($p)))
			{
				$pathes[] = $p;
			}
		}
		$subscriptions = array();
		if (($all_props = Vfs::propfind($pathes, self::SUBSCRIPTION_NS)))
		{
			foreach($all_props as $path => $props)
			{
				foreach($props as $prop)
				{
					if ($prop['name'] == self::SUBSCRIPTIONS_ATTR && $prop['val'])
					{
						$subscriptions[$path] = array();
						foreach(explode('::', $prop['val']) as $val)
						{
							$subscriptions[$path][] = $val;
						}
					}
				}
			}
		}
		return $parents_too ? $subscriptions : $subscriptions[$path];
	}

	/**
	 * Set subscription for a given path and user(s)
	 *
	 * @param string $path
	 * @param int|string|array $who account_id or group or email to add a single subscription, or array of account_id[:email|group] to set all subscriptions
	 * @param boolean $ignore_acl =false true check acl: read-rights to path for user, and users can only subscribe themselfs, admins also groups
	 * @return boolean true on success, false on error
	 */
	public static function set_subscription($path, $who, $ignore_acl=false)
	{
		// for storing everything but a user, get's prefixed with account_id of user subscribing it (separated by a colon)
		if (!is_array($who) && $who != $GLOBALS['egw_info']['user']['account_id'])
		{
			$who = $GLOBALS['egw_info']['user']['account_id'].($who?':'.$who:'');
		}
		if (!$ignore_acl && (is_array($who) || !self::allow_subscribe($who)))
		{
			if (self::DEBUG) error_log(__METHOD__."('$path', ".array2string($who).', FALSE) returning FALSE - permission denied');
			return false;	// permission denied
		}
		if (!is_array($who) && ($current = self::get_subscriptions($path, false)))
		{
			if (!in_array($who, $current)) $current[] = $who;
			$who = $current;
		}
		$ret = Vfs::proppatch($path, array(array(
			'ns' => self::SUBSCRIPTION_NS,
			'name' => self::SUBSCRIPTIONS_ATTR,
			'val' => $who ? implode('::', (array)$who) : null,
		)));
		if (self::DEBUG) error_log(__METHOD__."('$path', ".array2string($who).', '.array2string($ignore_acl).') returning '.array2string($ret));
		return $ret;
	}

	/**
	 * Delete subscription for a given path and user or all users
	 *
	 * @param string $path
	 * @param int|string $who account_id[:email|group] or null for all users
	 * @param boolean $ignore_acl =false true check acl: read-rights to path for user, and users can only subscribe themselfs, admins also groups
	 * @return boolean true if subscription was deleted, false if not (eg. because it was never set or $path does not exist)
	 */
	public static function delete_subscription($path, $who, $ignore_acl=false)
	{
		if (!$ignore_acl && (!$who || !self::allow_subscribe($who)))
		{
			if (self::DEBUG) error_log(__METHOD__."('$path', ".array2string($who).', FALSE) returning FALSE - permission denied');
			return false;	// permission denied
		}
		if ($who)
		{
			// for storing everything but a user, get's prefixed with account_id of user subscribing it (separated by a colon)
			if ($who != $GLOBALS['egw_info']['user']['account_id']) $who = $GLOBALS['egw_info']['user']['account_id'].($who?':'.$who:'');

			$current = self::get_subscriptions($path, false);
			if (($key = array_search($who, $current)) === false)
			{
				if (self::DEBUG) error_log(__METHOD__."('$path', ".array2string($who).', '.array2string($ignore_acl).') NOT in '.array2string($current).' returning FALSE');
				return false;	// user not set
			}
			unset($current[$key]);
		}
		else
		{
			$current = array();	// unset all
		}
		$ret = self::set_subscription($path, $current, true);	// true = no further acl check
		if (self::DEBUG) error_log(__METHOD__."('$path', ".array2string($who).', '.array2string($ignore_acl).') returning '.array2string($ret));
		return $ret;
	}

	/**
	 * Checks if current user is allowed to subscribe user or group $account_id
	 *
	 * @param int|string $who account_id or email
	 * @return boolean true if user is allowed, false if not
	 */
	public static function allow_subscribe($who)
	{
		if (strpos($who, ':') !== false)
		{
			list($owner, $who) = explode(':', $who);
		}
		else
		{
			$owner = $GLOBALS['egw_info']['user']['account_id'];
		}
		if (Vfs::$is_root)
		{
			$ret = true;	// root can eg. delete all subscribtions
		}
		elseif (!is_numeric($who) && strpos($who, '@'))
		{
			$ret = $owner == $GLOBALS['egw_info']['user']['account_id'];	// everyone is allowed to (un)subscribe his own email addresses
		}
		else
		{
			switch ($type = $GLOBALS['egw']->accounts->get_type($who))	// cast to int to get ride optional :email
			{
				case 'u':
					$ret = (int)$who === (int)$GLOBALS['egw_info']['user']['account_id'];
					break;

				case 'g':
					$ret = Vfs::$is_admin;	// todo group-admins
					break;

				default:
					$ret = false;
					break;
			}
		}
		if (self::DEBUG) error_log(__METHOD__."('$who') owner=$owner, type='$type' returning ".array2string($ret));
		return $ret;
	}
}
