<?php
/**
 * EPL Functions: Infolog Calendar Integration
 *
 * @link http://www.egroupware.org
 * @package epl
 * @subpackage infolog
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2010-16 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

use EGroupware\Api;
use EGroupware\Api\Framework;

/**
 * Infolog Calendar Integration
 *
 * Display infolog entries in calendar
 */
class stylite_infolog_calendar_integration
{
	const APP_NAME = 'infolog';
	const TABLE_NAME = 'egw_infolog';
	const USERS_TABLE = 'egw_infolog_users';

	protected static $filter_map = array(
		'default'	=>	'open',
		'all'		=>	'none',
		'deleted'	=>	'deleted',
	);

	/**
	 * Moves an integrated infolog event to another date/time in calendar views
	 * -Used by drag and drop in calendar app
	 *
	 * @param string $infolog_id infolog id
	 * @param string $targetDateTime date/time the entry should be moved to
	 * @param string $durationT = null amount of time that the entry's time should be increased/decreased by resize action
	 */
	static public function ajax_moveInfologEvent($infolog_id, $targetDateTime, $durationT = null)
	{
		$infolog_bo = new infolog_bo();

		$response = Api\Json\Response::get();

		$infolog = $infolog_bo->read($infolog_id);

		if (!$durationT && $infolog['info_enddate'])
		{
			$duration=$infolog['info_enddate'] - $infolog['info_startdate'];
		}
		else
		{
			$duration = $durationT;
		}

		$infolog['info_startdate'] = calendar_bo::date2ts($targetDateTime);
		if ($duration) $infolog['info_enddate'] = $infolog['info_startdate'] + $duration;
		if (is_array($infolog))
		{
			if (($res = $infolog_bo->write($infolog, true, true, true, $infolog['no_notifications'])))
			{
				// Directly update stored data.  If infolog is still in nextmatch, it will
				// be notified & update itself.
				$info = $infolog_bo->read($infolog_id);
				$response->call('egw.dataStoreUID','infolog::'.$infolog_id,$info);

				// Update calendar too
				$GLOBALS['egw_info']['flags']['currentapp'] = 'calendar';
				$cal = new calendar_ui();
				$infos = $cal->bo->search(array('query'=>$infolog_id,'offset' => 0));
				foreach($infos as $info)
				{
					$cal->to_client($info);
					$response->call('egw.dataStoreUID','calendar::'.$info['id'],$info);
				}
			}
			else
			{
				Framework::message('Failed to edit this integrated infolog event!');
			}
		}
	}

	/**
	 * Return query for calendar's search union query
	 *
	 * @param array $data
	 * @param string $data['cols'] cols query is suppost to return to match calendar query
	 * @param int $data['start'] start-date ts in servertime
	 * @param int $data['end'] end-date ts in servertime
	 * @param int|array $data['users'] users for which to return data (incl. all members and memberships as used in calendar_so)
	 * @param int|array $data['users_raw'] users for which to return data
	 * @param int $data['cat_id']
	 * @param string $data[	'filter'] filter-name: all (not rejected), accepted, unknown, tentative, rejected
	 * @param string $data[	'query']' pattern so search for, if unset or empty all matching entries are returned (no search)
	 * @return array values for keys 'selectes', ...
	 */
	static public function calendar_search_union(array $data)
	{
		//error_log( __METHOD__ . '() '.array2string($data));
		$pref = $GLOBALS['egw_info']['user']['preferences']['infolog']['calendar_integration'];

		if (empty($pref)) return null;

		$infolog = new infolog_bo();

		// some variables to simplify using SQL returned by our DB abstraction
		$startdate = $GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime('info_startdate'),'%Y-%m-%d');
		$starttime = $GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime('info_startdate'),'%H:%i:%s');
		$enddate = $GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime('info_enddate'),'%Y-%m-%d');

		$app_cols = array(
			'cal_id' => $GLOBALS['egw']->db->concat("'".self::APP_NAME."'",$GLOBALS['egw']->db->to_varchar(self::TABLE_NAME.'.info_id')),
			'cal_title' => 'info_subject',
			'cal_description' => 'info_des',
			// Fancy date processing to handle multi-day infologs.  If start and end are on different days, make start date 00:00, end date 23:59:59
			'cal_start' => "CASE WHEN info_enddate = 0 OR info_enddate < info_startdate OR $startdate = $enddate ".
				'THEN info_startdate ELSE '.$GLOBALS['egw']->db->unix_timestamp($startdate).' END',
			'cal_end' => "CASE WHEN info_enddate = 0 AND $starttime = '00:00:00' THEN info_startdate+86400-1 ".
				'WHEN info_enddate = 0 OR info_enddate < info_startdate THEN info_startdate+3600 '.
				"WHEN $startdate = $enddate THEN info_enddate ".
				'ELSE '.$GLOBALS['egw']->db->unix_timestamp($enddate).'+86400-1 END',
			'cal_owner' => 'info_owner',
			'cal_non_blocking' => 1,
			'cal_public' => "CASE info_access WHEN 'public' THEN 1 ELSE 0 END",
			'cal_uid' => 'info_uid',
			'cal_etag' => 'info_datemodified',
			'cal_location' => 'info_location',
			'cal_category' => $GLOBALS['egw']->db->to_varchar('info_cat'),
			'tz_id' => calendar_timezones::tz2id($GLOBALS['egw_info']['server']['server_timezone']),
			'cal_modified' => 'info_datemodified',
			'cal_modifier' => 'info_modifier',
			'cal_recurrence' => 0,
			'cal_priority' => "CASE info_priority WHEN 3 THEN 3 ELSE info_priority+1 END",
			'participants' => 'COALESCE('.$GLOBALS['egw']->db->group_concat('attendees.account_id').',info_owner)',
			'icons' => $GLOBALS['egw']->db->concat("'".self::APP_NAME.":'",'info_type'),
		);
		$where = array('info_type' => explode(',',$pref));
		// we need to obey infologs ACL!
		$users = array();
		foreach((array)$data['users_raw'] as $uid)
		{
			if (is_numeric($uid)) $users[] = $uid;
		}
		if (!$users) return null;	// no real users selected --> no infolog data for them
		$where[] = $acl_filter=$infolog->so->aclFilter('user'.implode(',',$users));
		//error_log("users_raw=".array2string($data['users_raw'])." --> ".array2string($users)." --> acl_filter='$acl_filter'");
		$status = self::$filter_map[$data['filter']] ? self::$filter_map[$data['filter']] : 'open';
		$where[] = $infolog->so->statusFilter($status,false);	// Show according to filter map
		if ($data['cat_id']) $where[] = $infolog->so->db->expression(self::TABLE_NAME, array('info_cat' => $data['cat_id']));

		// we use startdate also for end, if no enddate is given
		$end = 'CASE WHEN info_enddate != 0 THEN info_enddate ELSE info_startdate END';
		if ($data['start']) $where[] = (int)$data['start'] . ' <= '.$end;
		if ($data['end']) $where[] = 'info_startdate <= '.(int)$data['end'];
		//error_log('data='.array2string($data).' --> where='.array2string($where));

		// search infolog
		if ($data['query'])
		{
			if(!is_array($data['query']))
			{
				$pattern = ' '.$GLOBALS['egw']->db->capabilities[Api\Db::CAPABILITY_CASE_INSENSITIV_LIKE].' '.$GLOBALS['egw']->db->quote('%'.$data['query'].'%');
				$columns = array('info_from','info_location','info_subject'/*,'info_extra_value'*/);
				// at the moment MaxDB 7.5 cant cast nor search text columns, it's suppost to change in 7.6
				if ($GLOBALS['egw']->db->capabilities['like_on_text']) $columns[] = 'info_des';

				$where[] = '('.(is_numeric($data['query']) ? $infolog->so->info_table.'.info_id='.(int)$data['query'].' OR ' : '').
					implode($pattern.' OR ',$columns).$pattern.') ';
			}
			else
			{
				foreach($data['query'] as $name => $value)
				{
					if (isset($app_cols[$name]))
					{
						$where[] = $app_cols[$name].' = '.$GLOBALS['egw']->db->quote($value);
					}
				}
			}
		}
		return array(
			'selects' => array(
				array(
					'table' => self::TABLE_NAME,
            		'join' => ' LEFT JOIN '.self::USERS_TABLE.' ON '.self::TABLE_NAME.'.info_id='.self::USERS_TABLE.'.info_id AND '.self::USERS_TABLE.'.info_res_deleted IS NULL'.
						' LEFT JOIN '.self::USERS_TABLE.' attendees ON '.self::TABLE_NAME.'.info_id=attendees.info_id AND attendees.info_res_deleted IS NULL',
            		'cols' => calendar_so::union_cols($app_cols,$data['cols']),
					'where' => $where,
					'app' => self::APP_NAME,
					'append' => 'GROUP BY '.self::TABLE_NAME.'.info_id',
				)
			),
			//'edit_link' => array(__CLASS__,'edit_link'),		// default use Link::edit
			//'is_private' => array(__CLASS__,'is_private'),	// default use Link::title
		);
	}

	/**
	 * Extra settings for infolog
	 *
	 * @param array $data
	 */
	static public function settings_infolog(array $data)
	{
		unset($data);	// not used, but required by function signature

		$info = new infolog_bo();

		return array(
			'calendar_integration' => array(
				'type'   => 'multiselect',
				'label'  => 'Which types should the calendar show like events?',
				'name'   => 'calendar_integration',
				'values' => array(0 => lang('None')) + $info->enums['type'],
				'help'   => 'These types are shown as events in the calendar grid for all views.',
				'xmlrpc' => True,
				'admin'  => False
			),
		);
	}
}
