<?php
/**
 * EPL Functions - PBX integration
 *
 * @link www.egroupware.org
 * @author rb(at)egroupware.org
 * @copyright (c) 2016 by rb(at)egroupware.org
 * @package epl
 */

use EGroupware\Api;

/**
 * EPL PBX integration
 *
 * Currently supported vendors:
 * - http://placetel.de/
 */
class stylite_pbx
{
	/**
	 * Methods callable via menuaction
	 */
	public $public_functions = array(
		'admin' => true,
		'call' => true,
	);

	const APP_NAME = 'stylite';
	const CACHE_TIME_SIP_UIDS = 900;

	/**
	 * Supported PBX types
	 *
	 * @var array vendor (as in pbx_type) => url
	 */
	static $api_urls = array(
		'placetel.de' => 'https://api.placetel.de/api/',
	);
	/**
	 * API key (stored as pbx_api_key in app config)
	 *
	 * @var string
	 */
	protected $api_key;
	/**
	 * API url (determined by pbx_type in app config)
	 *
	 * @var string
	 */
	protected $api_url;

	/**
	 * Debug level set in config
	 *
	 * @var string null, "verbose" or "debug"
	 */
	protected $debug;

	/**
	 * Constructor
	 */
	function __construct()
	{
		$config = Api\Config::read(self::APP_NAME);
		$this->api_key = $config['pbx_api_key'];
		$this->api_url = self::$api_urls[$config['pbx_type']];
		$this->debug = $config['pbx_debug'];
	}

	/**
	 * Initiate a call to number send via _GET parameter "target"
	 */
	function call()
	{
		Api\Header\Content::type('', 'text/plain', 0, true, false);

		try {
			$result = $this->api_call('initiateCall.json', array(
				'sipuid' => $this->sip_uid(),
				'target' => $_GET['target'],
			));
			if ($this->debug)
			{
				echo $result['descr'];
			}
			else
			{
				header("HTTP/1.1 204 No Content");
			}
		}
		catch(Api\Exception $e) {
			_egw_log_exception($e);
			if ($this->debug)
			{
				echo $e->getMessage();
			}
			else
			{
				throw $e;
			}
		}
		exit;
	}

	/**
	 * Get SIP user for current user
	 */
	protected function sip_uid()
	{
		//Api\Cache::unsetInstance(__CLASS__, 'sip_uids');
		$users = Api\Cache::getInstance(__CLASS__, 'sip_uids', function()
		{
			$users = array();
			foreach($this->api_call('getVoIPUsers.json') as $data)
			{
				$users[$data['name']] = $data['uid'];
			}
			return $users;
		}, array(), self::CACHE_TIME_SIP_UIDS);

		$uid = $users[$GLOBALS['egw_info']['user']['account_lid']];
		if (empty($uid) && empty($uid=$users[$GLOBALS['egw_info']['user']['account_fullname']]))
		{
			throw new Api\Exception\WrongParameter("No sip uid for user '{$GLOBALS['egw_info']['user']['account_lid']}' or '{$GLOBALS['egw_info']['user']['account_fullname']}' found!");
		}
		//error_log(__METHOD__."() returning '$uid', users=".array2string($users));
		return $uid;
	}

	/**
	 * Call placetel.de api
	 *
	 * @param string $method method string eg. "iniciateCall.json" incl. ".json" prefix
	 * @param array $params =array() parameters without api_key, which get added automatic
	 */
	protected function api_call($method, $params=array())
	{
		$params['api_key'] = $this->api_key;
		//error_log(__METHOD__."('$method', ".array2string($params).')');

		if (($data = file_get_contents($this->api_url.$method, false, Api\Framework::proxy_context(null, null, array(
			'header'  => "Content-type: application/x-www-form-urlencoded\r\n",
			'method'  => 'POST',
			'content' => http_build_query($params),
		)))) === false)
		{
			if (!ini_get('allow_url_fopen'))
			{
				throw new Api\Exception\NoPermission(lang('%1 setting "%2" = %3 disallows access via http!',
					'php.ini', 'allow_url_fopen', array2string(ini_get('allow_url_fopen'))));
			}
			throw new Api\Exception\NoPermission('Webserver returned: '.$http_response_header[0]);
		}
		$data = json_decode($data, true);
		if (!$data || !is_array($data))
		{
			throw new Api\Exception('Invalid data returned!');
		}
		if (!empty($data['result']) && $data['result'] === '-1')
		{
			throw new Api\Exception("Placetel.de API call '$method' failed: $data[descr] ($data[error_code])");
		}
		return $data;
	}

	/**
	 * Register site config validation hooks
	 */
	public static function config_validate()
	{
		$GLOBALS['egw_info']['server']['found_validation_hook'] = array('final_validation');
	}

	/**
	 * Set telephony integration to this class call method
	 *
	 * @param array $config
	 */
	public static function configure_pbx_type($config)
	{
		if (isset(self::$api_urls[$config['pbx_type']]))
		{
			if (empty($config['pbx_api_key']))
			{
				$GLOBALS['config_error'] = 'API key must not be empty!';
				return;
			}
			config::save_value('call_link', Api\Egw::link('/index.php', 'menuaction=stylite.stylite_pbx.call&target=').'%1', 'phpgwapi');
			config::save_value('call_popup', $config['pbx_debug'] ? '480x200' : '_self', 'phpgwapi');
		}
	}
}

// Not part of the class, since config hooks are still using the old style
function final_validation($config)
{
	stylite_pbx::configure_pbx_type($config);
}
