<?php
/**
 * EPL Functions: Spam Titan API
 *
 * @link http://www.egroupware.org
 * @package epl
 * @subpackage api
 * @author Hadi Nategh <hn@egroupware.org>
 * @author Ralf Becker <rb@egroupware.org>
 * @copyright (c) 2016-18 by EGroupware GmbH
 */

class stylite_spamtitan
{
	/**
	 * default egroupware's SpamTitan API Url
	 */
	const DEFAULT_API_URL = 'https://mailgateway.egroupware.org';

	/**
	 * Default
	 * @var type
	 */
	private $data = array();

	/**
	 * Enable logging to error_log
	 */
	const LOG = false;

	function __construct(array $data=array())
	{
		$this->data = array_merge(array(
			'api_url'	=> self::DEFAULT_API_URL,
			'server'	=> '10.44.88.130',
			'imap'		=> 'imap.egroupware.org',
			'imap_port'	=> 993
		), $data);
	}
	 /**
	 * Make a call to SpamTitan API
	 *
	 * @param string $path path of api-url eg. '/domaingroup/add'
	 * @param array $_params get parameters for api-call
	 * @return SimpleXMLElement unserialized XML response
	 * @throws Exception if no valid response is received from SpamTitan
	 *
	 * @link http://download.spamtitan.com/manuals/SpamTitanAPIGuideV2.pdf
	 */
	public function api_call($path, array $_params)
	{
		$url = !empty($this->data['api_url']) ? $this->data['api_url'] : self::DEFAULT_API_URL;
		$url .= $path;

		foreach($_params as $name => &$value)
		{
			$value = urlencode($name).'='.strtr(urlencode($value), array(
				// Spamtitian does not decode these :(
				'%7B' => '{',
				'%3A' => ':',
				'%7D' => '}',
			));
		}
		$url .= '?'.implode('&', $_params);

		$context  = stream_context_create(array(
			'http' =>
				array(
					// SpamTitan requires "User-Agent: stoa" or API access for EGroupware IP!
					'user_agent' => 'stoa',
				)
		));
		if (!($xml = file_get_contents($url, false, $context)))
		{
			throw new Exception("Error contacting SpamTitan: $url");
		}
		return new SimpleXMLElement($xml);
	}

	/**
	 * Authenticate user AND create a user policy (if not existing)
	 *
	 * This method is neccessary, as without user policy, white/blacklist or spam/ham fails!
	 *
	 * Creating or querying user policy directly is an API operation and requires
	 * EGroupware server IP to be added to SpamTitan, which give access to everything.
	 *
	 * @param string $user
	 * @param string $passwd
	 * @return boolean|string true on success, false network error, string with error-message
	 */
	public function auth_user($user, $passwd)
	{
		$url = !empty($this->data['api_url']) ? $this->data['api_url'] : self::DEFAULT_API_URL;
		$url .= '/login.php';

		$matches = null;
		/* seems not necessary to create a session first (leave the code in case it will be required in future)
		if (!($get_response = file_get_contents($url)))
		{
			error_log(__METHOD__."('$user', \$passwd) failed to connect $url!");
			return false;
		}
		//var_dump($http_response_header);
		if (!preg_match('/^Set-Cookie:\s*([^;]+);/m', implode("\n", $http_response_header), $matches))
		{
			error_log(__METHOD__."('$user', \$passwd) failed to get session-cookie from $url!");
			return false;
		}
		//var_dump(preg_match_all('/<input[^>]+>/', $get_response, $matches2), $matches2);*/
		$opts = array('http' =>
			array(
				'method'  => 'POST',
				'header'  => array(
					'Content-Type: application/x-www-form-urlencoded',
					//'Cookie: '.$matches[1],
					//'User-Agent: EGroupware '.$GLOBALS['egw_info']['server']['api']['version'],
					//'Referer: '.$url,
				),
				'content' => http_build_query(
					array(
						//'jaction' => 'none',
						'address' => $user,
						'passwd' => $passwd,
					)
				)
			)
		);

		if (!($response = file_get_contents($url, false, stream_context_create($opts))))
		{
			error_log(__METHOD__."('$user', \$passwd) failed to connect $url!");
			return false;
		}
		if (preg_match('|<td colspan="3" class="error"><span id="login_status_Msg">([^<]+)</span></td>|', $response, $matches))
		{
			error_log(__METHOD__."('$user', $passwd) failed to login: $matches[1]");
			return $matches[1];
		}
		return true;
	}

	/**
	 * Learn email as Spam or Ham
	 *
	 * @param string $_mailbody Required. Email content
	 * @param string $_type = Spam, type of message
	 *		- Types: Spam | Ham
	 * @return string|boolean returns true if learning has been successful otherwise false
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function learnmail($_mailbody, $_type = 'Spam')
	{
		$url = (!empty($this->data['api_url'])? $this->data['api_url']: self::DEFAULT_API_URL).'/bayes/learnmail/'.
				'?user='.$this->data['user'].'&userpwd='.$this->data['userpwd'];
		$postdata = http_build_query(
			array(
				'mail_body' => urlencode($_mailbody),
				'type' => $_type
			)
		);
		$opts = array('http' =>
			array(
				'method'  => 'POST',
				'header'  => 'Content-type: application/x-www-form-urlencoded',
				// SpamTitan requires "User-Agent: stoa" or API access for EGroupware IP!
				'user_agent' => 'stoa',
				'content' => $postdata
			)
		);
		if (!($xml = file_get_contents($url, false, stream_context_create($opts))))
		{
			if (self::LOG) error_log(__METHOD__."(".substr($_mailbody, 0, 100)."... , '$_type') error contacting SpamTitian url=$url");
			throw new Exception("Error contacting SpamTitan");
		}
		$resp =  new SimpleXMLElement($xml);
		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."(".substr($_mailbody, 0, 100)."... , '$_type') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * List all the domains relayed by SpamTitan
	 *
	 * @param string $_server Optional. List Domain’s based on the IP or FQDN.
	 * @return array of domain-names or empty array if server is not found
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function list_domains($_server)
	{
		$resp = $this->api_call('/domain/list', array('server' => $_server));

		$ret = $resp['stat'] == 'ok' ? (array)$resp->Domains->Domain : array();
		if (self::LOG) error_log(__METHOD__."('$_server') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Adds the specified domain to the current list of domains relayed
	 * by SpamTitan to the specified destination server.
	 *
	 * @param string $_name Required. The name of the Domain to
	 *  add. MX Lookup is automatically enabled for new domains.
	 *
	 * @param type $_server Required. The destination mail server
	 *  for this domain. To specify a non-default port
	 *  (25) append ":8025".
	 *	- To specify a fallback server enter a list of IP addresses/FQDN’s separated by spaces.
	 *		SpamTitan will attempt delivery in the order listed. E.g.: “server=1.2.3.4 5.6.2.3”
	 *	- Specify a comma (,) separated list of IP addresses and/or FQDN’s. In this
	 *		case SpamTitan will attempt delivery in a round-robin fashion between the
	 *		listed servers. E.g.:”server=1.2.3.4,1.2.3.5”
	 * @param string $_domaingroup Optional. The Domain Group to add this domain to
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function add_domain ($_name, $_server, $_domaingroup)
	{
		$resp = $this->api_call('/domain/add', array('name' => $_name,'server' => $_server, 'domaingroup' => $_domaingroup));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_name', '$_server', '$_domaingroup') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Tests if the specified domain is been relayed by SpamTitan
	 *
	 * @param string $_name Required. The name of the Domain to add
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function domain_exists ($_name)
	{
		$resp = $this->api_call('/domain/exists', array('name' => $_name));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_name') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Returns extended information of a given domain, specified by
	 * domain name as per the required name parameter below.
	 *
	 * @param string $_name Required. The name of the Domain to add
	 * @return array return information regarding the requested domain
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function show_domain ($_name)
	{
		$resp = $this->api_call('/domain/show', array('name' => $_name));

		$ret = $resp['stat'] == 'ok' ? $resp->Data : array();
		if (self::LOG) error_log(__METHOD__."('$_name') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Delete a domain. Note that this will delete all user policies
	 * associated with this domain.
	 *
	 * @param string $_name Required. The name of the domain to delete.
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function delete_domain ($_name)
	{
		$resp = $this->api_call('/domain/delete', array('name' => $_name));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_name') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Tests if the specified email address/domain is blacklisted
	 * by the specified user/domain, or globally if no user/domain is specified.
	 *
	 * @param string $_sender Required. The email address or domain to
	 *  check for the existence of in a particular blacklist
	 *
	 * @param string $_user Optional. The email address or domain name
	 *  of the blacklist to check. If no user is specified, then the
	 *  global blacklist will be checked
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function check_blacklist ($_sender, $_user)
	{
		$resp = $this->api_call('/blacklist/exists', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Add an email address or domain to a specified users or domains
	 * blacklist. If no user/domain is specified then the entry is added
	 * to the global blacklist.
	 *
	 * @param string $_sender Required. The email address or domain to
	 *  add to the blacklist
	 *
	 * @param string $_user Optional. The email address or domain name
	 *  of the blacklist entry to add. If no user is specified, then the
	 *  entry will be added to the global blacklist.
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function add_blacklist_entry ($_sender, $_user)
	{
		$resp = $this->api_call('/blacklist/add', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Delete an entry from the specified users/domains blacklist.
	 * If no user is specified, then delete the entry from the global blacklist.
	 *
	 * @param string $_sender Required. The email address or domain
	 * to delete from the blacklist
	 *
	 * @param string $_user Optional. The email address or domain name
	 * of the owner of the blacklist entry to delete. If no user is specified,
	 * then the entry will be deleted from the global blacklist.
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function delete_blacklist_entry ($_sender, $_user)
	{
		$resp = $this->api_call('/blacklist/delete', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * List all Whitelisted email addresses and/or domains for the
	 * specified user/domain. If no user or domain is specified then
	 * the global Whitelist will be returned.
	 *
	 * @param string $_user Optional. The email address or domain name
	 * of the user/domain for which to view the whitelist
	 *
	 * @return array return success response or false
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function show_whitelist ($_user)
	{
		$resp = $this->api_call('/whitelist/exists', array('user' => $_user));

		$ret = $resp['stat'] == 'ok' ? $resp->Whitelist : array();
		if (self::LOG) error_log(__METHOD__."('$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Tests if the specified email address/domain is whitelisted by
	 * the specified user/domain, or globally if no user/domain is specified.
	 *
	 * @param string $_sender Required. The email address or domain
	 * to check for the existence of in a particular whitelist
	 *
	 * @param string $_user Optional.  The email address or domain name
	 * of the whitelist to check. If no user is specified, then the global
	 * whitelist will be checked
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function check_whitelist ($_sender, $_user)
	{
		$resp = $this->api_call('/whitelist/exists', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Add an email address or domain to a specified users or domains
	 * whitelist. If no user/domain is specified then the entry is
	 * added to the global whitelist.
	 *
	 * @param string $_sender Required. The email address or domain to
	 * add to the whitelist
	 *
	 * @param string $_user Optional. The email address or domain name
	 *  of the whitelist entry to add. If no user is specified, then the
	 *  entry will be added to the global whitelist.
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function add_whitelist_entry ($_sender, $_user)
	{
		if (!empty($_user)) $this->data['user'] = $_user;
		$resp = $this->api_call('/whitelist/add', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Delete an entry from the specified users/domains whitelist.
	 * If no user is specified, then delete the entry from the global whitelist.
	 *
	 * @param string $_sender Required. The email address or domain
	 * to delete from the whitelist
	 *
	 * @param string $_user Optional. The email address or domain name
	 * of the owner of the whitelist entry to delete. If no user is specified,
	 * then the entry will be deleted from the global whitelist.
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function delete_whitelist_entry ($_sender, $_user)
	{
		$resp = $this->api_call('/whitelist/delete', array('sender' => $_sender, 'user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_sender', '$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Add user policy. The policy which is added will be the same policy as the
	 * domain policy. It can subsequently be modified using the edit API.
	 *
	 * @param string $_name Required. The email address of the policy to add
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function add_policy ($_name)
	{
		$resp = $this->api_call('/policy/add', array('name' => $_name));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_name') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Returns if the specified email address or domain policy exists.
	 *
	 * @param string $_name Required. The email address or domain of the policy
	 * to check for.
	 *
	 * @return boolean|string return true if successful otherwise error message
	 * @throws Exception if no valid response is received from SpamTitan
	 */
	public function policy_exists ($_name)
	{
		$resp = $this->api_call('/policy/exists', array('name' => $_name));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_name') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Returns aliases of given user/email or primary user for an alias
	 *
	 * @param string $_user email address
	 * @return array|string values for keys "user", "aliases", "is_primary", "is_secondary" and "my_primary" or string with error-message
	 *  "No such alias" is returned, if there is not yet an alias!
	 */
	public function show_alias($_user)
	{
		$resp = $this->api_call('/alias/show', array('user' => $_user));

		$aliases = array();
		if ($resp['stat'] == 'ok' && $resp->Aliases)
		{
			foreach($resp->Aliases->Status as $status)
			{
				$aliases[] = (string)$status->alias;
			}
		}
		$ret = $resp['stat'] == 'ok' ? array(
			'user' => (string)$resp->User,
			'aliases' => $aliases,
			'is_primary' => (boolean)$resp->Is_Primary,
			'is_secondary' => (boolean)$resp->Is_Secondary,
			'my_primary' => (string)$resp->My_Primary,
		) : (string)(string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Returns all aliases of a given (main) account / email
	 *
	 * @param string $_user main email address
	 * @return int|string 0: no user or alias, 1: user, 2: alias, 3: user and alias or string with error-message
	 */
	public function alias_exists($_user)
	{
		$resp = $this->api_call('/alias/exists', array('user' => $_user));

		$ret = $resp['stat'] == 'ok' ? ($resp->Is_Primary ? 1 : 0)+($resp->Is_Secondary ? 2 : 0) : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_user') returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Returns all aliases of a given (main) account / email
	 *
	 * @param string $_user main email address
	 * @param string|array $_aliases alias(es) to add
	 * @return boolean|string true on success or string with error-message
	 */
	public function add_alias($_user, array $_aliases)
	{
		$resp = $this->api_call('/alias/add', array('user' => $_user, 'secondary' => implode(',', (array)$_aliases)));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_user', ".array2string($_aliases).") returning ".array2string($ret));
		return $ret;
	}

	/**
	 * Remove an alias
	 *
	 * @param string $_user email address or alias to remove
	 * @return boolean|string true on success or string with error-message
	 */
	public function delete_alias($_user)
	{
		$resp = $this->api_call('/alias/delete', array('user' => $_user));

		$ret = $resp['stat'] == 'ok' ? true : (string)$resp->error;
		if (self::LOG) error_log(__METHOD__."('$_user') returning ".array2string($ret));
		return $ret;
	}
}
