#!/usr/bin/env php
<?php
/**
 * Migration of Contacts, linked InfoLog and Timesheet entries plus their files
 *
 * This script does NOT export/import accounts, ACL or custom-fields, this has to be done manual before.
 * ACL and accounts can be exported by explicit using appname "acl" and/or "accounts" in --export!
 * You have to create a JSON file specifying conversation of account_id's and
 * pass it to: migrate --import <json-to-import> --exported <account-conversation>:
 *
 * {"accounts": { "<old-account-id>": <new-account-id>, ...[, "0": <default-id-for-not-specified-accounts>] }}
 *
 * Above JSON file can also be used to specify an explicit conversation for category-ids,
 * which otherwise will be exported automatic (use "categories": {"<old-cat-id>": <new-cat-id>, ...})
 *
 * Examples:
 * sudo -u www bash -c "php -dmemory_limit=-1 stylite/migrate.php --export list=32 --instance ralfsmac.local --debug | stylite/migrate.php --import /var/lib/egroupware/default/files --instance test.dev"
 *
 * php -dmemory_limit=-1 stylite/migrate.php --instance stylite.de \
 *	--export list=313 > ~/export-egw-stylite.json
 * php -dmemory_limit=-1 stylite/migrate.php --instance stylite.de \
 *	--export addressbook,acl,calendar > ~/export-egw-ab-acl-cal.json
 * php -dmemory_limit=-1 stylite/migrate.php --instance pole.egroupware.net \
 *	--import /var/lib/egroupware/stylite.de/files --exported /home/dev/stylite2egw.json \
 *	/home/dev/export-egw-stylite.json 2>&1 > ~/imported-pole.json | tee ~/imported-pole.log
 *
 * @link http://www.egroupware.org
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2016/17 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

use EGroupware\Api;

if (($key = array_search('--instance', $_SERVER['argv'])) !== false && $_SERVER['argc'] > $key+1)
{
	$_REQUEST['domain'] = $_SERVER['argv'][$key+1];
	unset($_SERVER['argv'][$key], $_SERVER['argv'][$key+1]);
	$_SERVER['argc'] -= 2;
}

$GLOBALS['egw_info'] = array(
	'flags' => array(
		'currentapp' => 'login',
	),
);

require __DIR__.'/../header.inc.php';

// stop all output buffering
for($n=0; $n < 10 && ob_get_level(); ++$n)
{
	ob_end_flush();
}

// do NOT run shutdown functions, this prevents a PHP Fatal Error on shutdown
define('EGW_SHUTDOWN', True);

/**
 * Old => new account-id(s)
 */
$accounts = array(
	0 => 5,		// sysop (default)
);

// give user all apps we want to export, as links are not reported otherwise
$GLOBALS['egw_info']['user']['apps'] =
$GLOBALS['egw_info']['apps'] = array(
	'addressbook' => true,
	'calendar' => true,
	'filemanager' => true,
	'infolog' => true,
	'timesheet' => true,
	'records' => true,
	'resources' => true,
	'admin' => true,
);

/**
 * app => array of ids
 */
$exported = array(
	'accounts' => $accounts,
	'addressbook' => array(),
	'calendar' => array(),
	'infolog' => array(),
	'timesheet' => array(),
	'records' => array(),
	'resources' => array(),
);
$linked = $attached = 0;

/**
 * Parsing command line arguments
 */
$cmd = array_shift($_SERVER['argv']); --$_SERVER['argc'];

$json_options = JSON_UNESCAPED_SLASHES|JSON_UNESCAPED_UNICODE;
if (($key = array_search('--debug', $_SERVER['argv'])) !== false)
{
	$json_options |= JSON_PRETTY_PRINT;
	unset($_SERVER['argv'][$key]);
	--$_SERVER['argc'];
}
$chunk_size = 500;
$max_level = 50;
$keep_account_id = $show_timing = false;

if (class_exists('records_bo'))
{
	$records_contact_cols = array(records_bo::CUSTOMER_COL, 'Organisation');
}

if (($key = array_search('--keep-account-id', $_SERVER['argv'])) !== false)
{
	$keep_account_id = true;
	unset($_SERVER['argv'][$key]);
	$_SERVER['argc'] -= 1;
}

if (($key = array_search('--show-timing', $_SERVER['argv'])) !== false)
{
	$show_timing = true;
	unset($_SERVER['argv'][$key]);
	$_SERVER['argc'] -= 1;
}

if (($key = array_search('--max-level', $_SERVER['argv'])) !== false && $_SERVER['argc'] > $key+1)
{
	$max_level = (int)$_SERVER['argv'][$key+1];
	unset($_SERVER['argv'][$key], $_SERVER['argv'][$key+1]);
	$_SERVER['argc'] -= 2;
}

if (($key = array_search('--exported', $_SERVER['argv'])) !== false && $_SERVER['argc'] > $key+1)
{
	$file = $_SERVER['argv'][$key+1];
	if (!file_exists($file) || !($exported = json_decode(file_get_contents($file), true)))
	{
		error_log("Already exported entries (--exported $file) file not found or no JSON!");
		exit(1);
	}
	if (empty($exported['accounts'])) $exported['accounts'] = $accounts;
	unset($_SERVER['argv'][$key], $_SERVER['argv'][$key+1]);
	$_SERVER['argc'] -= 2;
}

if (($key = array_search('--export', $_SERVER['argv'])) !== false && $_SERVER['argc'] == 2)
{
	export($_SERVER['argv'][1+$key]);
	exit;
}

if (($key = array_search('--import', $_SERVER['argv'])) !== false && $_SERVER['argc'] >= 2)
{
	$exported_files_dir = $_SERVER['argv'][1+$key];
	if (file_exists($exported_files_dir) && is_dir($exported_files_dir))
	{
		$input_file = $_SERVER['argc'] == 3 ? array_pop($_SERVER['argv']) : null;
		if (!empty($input_file) && file_exists($input_file))
		{
			$source = fopen($input_file, 'r');
		}
		else
		{
			$source = STDIN;
		}
		import($source);
		exit;
	}
}

if ($_SERVER['argv'])
{
	echo "\nUnknown options: ".implode(' ', $_SERVER['argv'])."\n";
}

echo "\nUsage:\t$cmd --export [(owner|list|cat_id)=]<ids-comma-sep>[,(<name>=<ids>),...] [--debug] [--max-level <N> (default 5)]\n";
echo "\t$cmd --export <app1>[,<app2>,...] [--debug] (currently supported: accounts, addressbook, calendar, infolog, timesheet, resources, filemanager, acl)\n";
echo "\t$cmd --import <files-dir-of-exporting-instance> [<file-to-import>] [> <imported-entries>]\n";
echo "\tCommon options for export and import:\n";
echo "\t--instance <name> instance to use\n";
echo "\t--exported <imported-entries> file with previously imported or exported entries and/or account conversation\n";
echo "\t--keep-account-id if no exported/changed account found, try for existing account with identical nummeric id\n";
echo "\t--show-timing some timing information (number of entries per app imported in what time)\n";
exit(1);

/*******************************************************************************
 * App unspecific export code
 ******************************************************************************/

/**
 * Export
 *
 * @param string $values [(owner|list|cat_id)=]<ids-comma-sep>[,(<name>=<ids>),...]
 */
function export($values)
{
	global $exported, $json_options, $max_level, $to_export, $chunk_size, $show_timing;

	$keys = $matches = array();
	while (preg_match('/^([a-z_]+)=([0-9,-]+),?(.*)/', $values, $matches))
	{
		$keys[$matches[1]] = array_diff(explode(',', $matches[2]), array(''));
		$values = $matches[3];
	}
	if (!$keys) $keys = array('list' => explode(',', $values));

	if (isset($keys['list']) && !is_numeric($keys['list'][0]))
	{
		$apps = $keys['list'];
	}

	if (php_sapi_name() != 'cli') header('Content-Type: text/plain');
	Api\Vfs::$is_root = true;

	echo "[\n";
	$to_export = array();

	if (isset($apps))
	{
		// allways export accounts first, as they are need for all apps
		if (($a = array_search('accounts', $apps)))
		{
			unset($apps[$a]);
			array_unshift($apps, 'accounts');
		}

		// allways export calendar after all other apps as it might need ids from addressbook and resources
		if (($a = array_search('calendar', $apps)) !== false)
		{
			unset($apps[$a]);
			$apps[] = 'calendar';
		}
		// allways export acl last, as it might use categories exported by apps
		if (($a = array_search('acl', $apps)) !== false)
		{
			unset($apps[$a]);
			$apps[] = 'acl';
		}

		foreach($apps as $app)
		{
			if (!is_callable('export_'.$app))
			{
				error_log("No method to export entries for app '$app' --> ignored!");
				continue;
			}
			$start = microtime(true);
			call_user_func('export_'.$app, array());
			if ($show_timing)
			{
				$num = count($exported[$app]);
				$time = microtime(true) - $start;
				$per_sec = round($num/$time);
				error_log("export_$app: $num $app entries exported in ".number_format($time,1)." secs --> $per_sec entries/sec");
			}
		}
	}
	else	// follow links of exported contacts
	{
		$level = 0;
		export_addressbook($keys);
		error_log("level $level: exported ".count($exported['addressbook'])." entries of addressbook (".array2string($keys).')');

		if (class_exists('records_bo'))
		{
			export_records($keys=array());
			error_log("level $level: exported ".count($exported['records'])." entries of records (".array2string($keys).')');
		}
		for($level=1; $to_export && $level < $max_level; ++$level)
		{
			foreach($to_export as $app => &$ids)
			{
				$num = 0;
				$ids = array_diff(array_unique($ids), array_keys((array)$exported[$app]));
				while($ids)
				{
					$ids_to_export = array_slice($ids, 0, $chunk_size);
					$ids = array_slice($ids, $chunk_size);
					call_user_func('export_'.$app, $ids_to_export);
					$num += count($ids_to_export);
				}
				unset($to_export[$app]);
				error_log("level $level: exported $num entries of $app");
			}
		}
		foreach($exported as $app => &$eids)
		{
			$eids = array_keys($eids);
		}
		echo json_encode($exported, $json_options);

		error_log("Followed links until level $level/$max_level.");
	}
	echo "]\n";

	report('exported');
}

/**
 * Export given entries
 *
 * @param array $entries
 * @param string $app
 * @param string $id_attr eg. "id"
 * @param string $cat_attr ='cat_id'
 */
function export_entries($entries, $app, $id_attr, $cat_attr='cat_id')
{
	global $exported, $json_options;

	if (!$entries || !is_array($entries)) return;

	$reindexed = array();
	foreach($entries as $entry)
	{
		if (!isset($exported[$app][$entry[$id_attr]]))
		{
			$entry['.app'] = $app;
			$id = $entry['.id'] = $entry[$id_attr];
			add_categories($entry, $cat_attr);
			// remove all null values
			$entry = array_diff($entry, array(null));
			$reindexed[$entry[$id_attr]] = $entry;
			$exported[$app][$id] = true;
		}
	}
	add_links($reindexed, $app);

	Api\Link::init_static();	// remove (title-)cache of Link class, to lower memory usage

	foreach($reindexed as $id => $entry)
	{
		echo json_encode($entry, $json_options).",\n";
	}
}

/**
 * Add catgories to export
 *
 * @param array $entry
 * @param string $name name of category attribute in $entry
 */
function add_categories(array &$entry, $name='cat_id')
{
	if (!empty($entry[$name]))
	{
		foreach(explode(',', $entry[$name]) as $cat_id)
		{
			export_cat($cat_id);
		}
	}
}

/**
 * Export category with all it's parents
 *
 * @param int $cat_id
 */
function export_cat($cat_id)
{
	global $exported, $json_options;

	if (!isset($exported['categories']) || !isset($exported['categories'][$cat_id]))
	{
		if ($cat_id > 0 && ($cat = Api\Categories::read($cat_id)))
		{
			if ($cat['parent']) export_cat($cat['parent']);

			$cat['.app'] = 'categories';
			$cat['.id'] = $cat_id;
			unset($cat['main'], $cat['level'], $cat['last_mod'], $cat['app_name']);
			echo json_encode($cat, $json_options).",\n";
			$exported['categories'][$cat_id] = true;
		}
	}
}

/**
 * Query and add links under key ".links" to entries to export and schedule linked entries for export
 *
 * @param array $entries $id => $entry pairs
 * @param string $app app-name entries are from
 */
function add_links(array &$entries, $app)
{
	global $to_export, $exported, $linked, $attached;

	foreach(Api\Link::get_links_multiple($app, array_keys($entries), false) as $id => $links)
	{
		if ($links)
		{
			foreach($links as $link_id => &$link)
			{
				if ($link['app'] == 'file')
				{
					if (!file_exists($path=Api\Vfs\Sqlfs\StreamWrapper::_fs_path(-$link_id)))
					{
						error_log("Attachment '$link[id]' to $app:$id NOT found in filesystem under $path!");
						unset($links[$link_id]);
						continue;
					}
					$link['path'] = substr($path, 1+strlen($GLOBALS['egw_info']['server']['files_dir']));
					unset($link['app2'], $link['id2']);
					++$attached;
				}
				else
				{
					if (!isset($exported[$link['app']][$link['id']]) &&
						!in_array($link['id'], (array)$to_export[$link['app']]))
					{
						$to_export[$link['app']][] = $link['id'];
					}
					++$linked;
				}
				unset($link['deleted'], $link['link_id']);
			}
			$entries[$id]['.links'] = $links;
		}
	}
}

/*******************************************************************************
 * App unspecific import code
 ******************************************************************************/

$links = array();

/**
 * Import entries as JSON from stdin
 *
 * @param resource $source content to import
 */
function import($source)
{
	global $exported, $json_options;

	$json = '';
	while(($line = fgets($source)) !== false)
	{
		if (empty($json) && $line[0] == '[')
		{
			if ($line == "[\n") continue;
			$line = substr($line, 1);
		}
		$json .= $line;
		if ($line == "},\n" || $line[0] == '{' && substr($line, -3) == "},\n")
		{
			if (!($entry = json_decode($json=substr($json, 0, -2), true)))
			{
				error_log("Could NOT decode JSON '$json'!");
				continue;
			}
			$json = '';
			import_entry($entry);
		}
	}
	import_links();

	report('imported');

	echo json_encode($exported, $json_options);
}

function report($msg)
{
	global $exported, $linked, $attached, $missing_accounts;

	$msgs = array();
	foreach($exported as $app => $entries)
	{
		if (!count($entries)) continue;
		$msgs[] = count($entries).' '.$app;
	}
	$msgs[] = $linked.' links';
	$msgs[] = $attached.' attachments';
	error_log(implode(', ', $msgs).' '.$msg.'.');

	if ($missing_accounts)
	{
		arsort($missing_accounts);
		error_log('Following accounts are missing (do not have a conversation) and how often say been used: '.array2string($missing_accounts));
	}
}

function import_links()
{
	global $exported, $links, $exported_files_dir, $linked, $attached;
	static $infolog_bo = null;

	Api\Vfs::$is_root = true;

	foreach($links as $app => $old_ids)
	{
		foreach($old_ids as $old_id => $all_links)
		{
			if (!isset($exported[$app][$old_id]))
			{
				error_log("$app:$old_id has not been imported, ignoring links!");
				continue;
			}
			$new_id = $exported[$app][$old_id];
			$update_infolog = array();

			foreach($all_links as $old_link_id => $link)
			{
				switch($old_link_id)
				{
					case 'info_id_parent':
						if (isset($exported['infolog'][$link]))
						{
							$update_infolog['info_id_parent'] = $exported['infolog'][$link];
						}
						break;
					case 'info_link_id':
						break;
					default:
						if ($old_link_id < 0)
						{
							if (!file_exists($path = $exported_files_dir.'/'.$link['path']))
							{
								error_log("$app:$old_id->$new_id: $path not found, ignoring attachment!");
							}
							elseif (!Api\Link::attach_file($app, $new_id, array(
								'name' => $link['id'],
								'type' => $link['type'],
								'tmp_name' => $path,
							), $link['remark']))
							{
								error_log("$app:$old_id->$new_id: error attaching $link[id] ($path)!");
							}
							else
							{
								++$attached;
							}
						}
						else
						{
							if (!isset($exported[$link['app']][$link['id']]))
							{
								error_log("$app:$old_id->$new_id: $link[app]:$link[id] has not been imported, ignoring links!");
								continue;
							}
							if (!($new_link_id = Api\Link::link($app, $new_id, $link['app'], $exported[$link['app']][$link['id']],
								$link['remark'], new_account_id($link['owner']), $link['lastmod'])))
							{
								error_log("$app:$old_id->$new_id: error linking with $link[app]:$link[id]->{$exported[$link['app']][$link['id']]}!");
								continue;
							}
							++$linked;

							if (isset($all_links['info_link_id']) && $all_links['info_link_id'] == $old_link_id)
							{
								$update_infolog['info_link_id'] = $new_link_id;
							}
						}
						break;
				}
			}
			if ($update_infolog)
			{
				if (!isset($infolog_bo)) $infolog_bo = new infolog_bo();
				if (($info = $infolog_bo->read($new_id, /*$run_link_id2from=*/false, /*$date_format=*/'ts', /*$ignore_acl=*/true)))
				{
					$info = array_merge($info, $update_infolog);
					$infolog_bo->write($info, /*$check_defaults=*/false, /*$touch_modified=*/false,
						/*$user2server=*/true, /*$skip_notification=*/true, /*$throw_exception=*/false,
						/*$purge_cfs=*/null, /*$ignore_acl=*/true);
				}
			}
		}
	}
}

function import_entry(array $entry)
{
	global $exported, $links;

	if (is_callable($func='import_'.$entry['.app']))
	{
		try {
			if (!empty($entry['.links']))
			{
				$links[$entry['.app']][$entry['.id']] = $entry['.links'];
			}
			//error_log(__METHOD__."() calling $func(".array2string($entry).")");
			$exported[$entry['.app']][$entry['.id']] = call_user_func($func, $entry);
			if (!$exported[$entry['.app']][$entry['.id']])
			{
				error_log("Error: $func could not import ".array2string($entry));
				unset($links[$entry['.app']][$entry['.id']]);
			}
		}
		catch (Exception $e) {
			error_log("Error: ".$e->getMessage().": $func could not import ".array2string($entry));
		}
	}
	else
	{
		error_log("Missing function $func to import ".array2string($entry));
	}
}

function import_categories(array $cat)
{
	unset($cat['.app'], $cat['.id'], $cat['id']);

	if ($cat['parent']) $cat['parent'] = new_cat_ids($cat['parent']);
	$cat['owner'] = $cat['owner'] ? (int)new_account_id($cat['owner']) : 0;

	$cats = new Api\Categories($cat['owner'], $cat['appname']);

	return $cats->add($cat);
}

$missing_accounts = array();

function new_account_id($old, $use_default=true)
{
	global $exported, $missing_accounts, $keep_account_id;

	if (isset($exported['accounts'][$old]))
	{
		return $exported['accounts'][$old];
	}
	// if --keep-account-id is used, check if account or group exists and use it
	if ($keep_account_id && $GLOBALS['egw']->accounts->get_type($old) === ($old < 0 ? 'g' : 'u'))
	{
		return $old;
	}
	if (!$use_default)
	{
		return null;
	}
	if (!isset($missing_accounts[$old]))
	{
		error_log("No new account_id for user #$old specified, using default #{$exported['accounts'][0]}!");
		$missing_accounts[$old] = 0;
	}
	++$missing_accounts[$old];

	return $exported['accounts'][0];
}

function new_cat_ids($old_ids)
{
	global $exported;

	$cats = is_array($old_ids) ? $old_ids : explode(',', $old_ids);
	foreach($cats as $n => $old_id)
	{
		if (isset($exported['categories'][$old_id]))
		{
			$cats[$n] = $exported['categories'][$old_id];
		}
		else
		{
			unset($cats[$n]);
		}
	}

	return is_array($old_ids) ? $cats : implode(',', $cats);
}

/*******************************************************************************
 * Accounts support
 ******************************************************************************/

/**
 * Export accounts
 *
 * @param array $keys
 */
function export_accounts($keys=array())
{
	global $chunk_size, $exported;

	$accounts = new Api\Accounts();
	$contacts = new Api\Contacts();

	// export groups without members first
	for($chunk=0;
		($entries = $accounts->search(array(
			'type' => 'groups',
			'start' => $chunk*$chunk_size,
			'offset' => $chunk_size,
			'order' => 'account_id',
			'sort'  => 'ASC',
		)+$keys));
		++$chunk)
	{
		foreach($entries as $account_id => &$entry)
		{
			unset($entry['account_status'], $entry['account_expires'], $entry['account_primary_group'], $entry['members']);
		}
		export_entries($entries, 'accounts', 'account_id', 'cat_id');
	}
	// then export accounts with memberships and contact-data
	for($chunk=0;
		($entries = $accounts->search(array(
			'type' => 'accounts',
			'start' => $chunk*$chunk_size,
			'offset' => $chunk_size,
			'order' => 'account_id',
			'sort'  => 'ASC',
		)+$keys));
		++$chunk)
	{
		$account_ids = array();
		foreach($entries as $account_id => &$entry)
		{
			if (!isset($entry['memberships'])) $entry['memberships'] = $accounts->memberships($account_id, true);
			$entry += $accounts->read($account_id);
			$account_ids[] = $account_id;
		}
		// add contact data for users incl. links
		if ($account_ids && ($users = $contacts->search(array(), false, 'egw_addressbook.contact_id ASC',
			'', '', false, 'AND', false, array('account_id' => $account_ids))))
		{
			$reindexed = array();
			foreach($users as &$entry)
			{
				if (!empty($entry['jpegphoto']))
				{
					$entry['jpegphoto'] = base64_encode($entry['jpegphoto']);
				}
				unset($entry['photo']);

				$reindexed[$entry['id']] = $entry;
			}
			// add custom fields
			foreach($contacts->read_customfields(array_keys($reindexed)) as $id => $cfs)
			{
				$reindexed[$id] += $cfs;
			}
			add_links($reindexed, 'addressbook');

			foreach($reindexed as &$entry)
			{
				$entries[$entry['account_id']] += $entry;

				// mark account-contact as exported
				$exported['addressbook'][$entry['id']] = true;
			}
		}
		export_entries($entries, 'accounts', 'account_id', 'cat_id');
	}
}

function import_accounts($account)
{
	global $exported, $links;

	static $accounts = null;
	if (!isset($accounts)) $accounts = new Api\Accounts();

	// check if account-name exists, they have to be unique!
	if (($account_id = $accounts->name2id($account['account_lid'], 'account_lid')))
	{
		// existing names will be reused, if the type (user/group) matches
		if (($account_id < 0) !== ($account['account_id'] < 0))
		{
			error_log("account '$account[account_lid]' (#$account[account_id]) already exists, but not of same type (user/group)!");
			return null;
		}
		error_log("account '$account[account_lid]' (#$account[account_id]) already exists as #$account_id --> using existing one unchanged!");
		return $account_id;
	}

	// if numerical id is taken, unset it to create a new one and not overwrite existing account
	if ($accounts->id2name($old_account_id = $account['account_id'])) unset($account['account_id']);

	// fix group-ids for users
	if ($old_account_id > 0)
	{
		$account['account_primary_group'] = new_account_id($account['account_primary_group'], false);
		foreach($account['memberships'] as &$gid)
		{
			$gid = new_account_id($gid, false);
		}
		// Accouns::save() uses account_passwd
		$account['account_passwd'] = $account['account_pwd'];
	}

	$account['account_id'] = $accounts->save($account);

	// deal with contact part of account
	if ($old_account_id > 0)
	{
		$accounts->set_memberships($account['memberships'], $account['account_id']);

		// we need to explicit add conversation for contact, as import_addressbook is not called by import_entry!
		$account['id'] = $exported['addressbook'][$account['id']] =  $GLOBALS['egw']->accounts->id2name($account['account_id'], 'person_id');
		import_addressbook($account, true);

		// links need to be handled by addressbook not accounts
		if (!empty($links['accounts'][$old_account_id]))
		{
			$links['addressbook'][$account['id']] = $links['accounts'][$old_account_id];
			unset($links['accounts'][$old_account_id]);
		}
	}
	return $account['account_id'];
}

/*******************************************************************************
 * Addressbook support
 ******************************************************************************/

/**
 * Export addresses matching given criteria
 *
 * @param array $keys
 */
function export_addressbook(array $keys)
{
	global $chunk_size;

	$contacts = new Api\Contacts();

	if (isset($keys[0])) $keys = array('id' => $keys, 'tid' => null);	// return deleted entries, if queried by id
	$keys['account_id'] = '';	// return deactivated accounts

	for($chunk=0;
		($entries = $contacts->search(array(), false, 'egw_addressbook.contact_id ASC',
			'', '', false, 'AND', array($chunk*$chunk_size, $chunk_size), $keys));
		++$chunk)
	{
		$id2key = array();
		foreach($entries as $key => &$entry)
		{
			if (!empty($entry['jpegphoto']))
			{
				$entry['jpegphoto'] = base64_encode($entry['jpegphoto']);
			}
			unset($entry['photo']);

			$id2key[$entry['id']] = $key;
		}
		// add custom fields
		foreach($contacts->read_customfields(array_keys($id2key)) as $id => $cfs)
		{
			$entries[$id2key[$id]] += $cfs;
		}
		export_entries($entries, 'addressbook', 'id', 'cat_id');
	}
}

function import_addressbook($contact, $update_account=false)
{
	static $contacts = null;
	if (!isset($contacts)) $contacts = new Api\Contacts();

	// handle accounts by using already existing accounts or
	// creating them as plain contacts in AB of default account
	if ($contact['account_id'] && !$update_account)
	{
		if (($new_account_id = new_account_id($contact['account_id'], false)) &&
			($new_id = $GLOBALS['egw']->accounts->id2name($new_account_id, 'person_id')))
		{
			return $new_id;
		}
		unset($contact['account_id']);
		$contact['owner'] = new_account_id($contact['account_id']);
	}

	foreach(array('owner', 'creator', 'modifier') as $name)
	{
		if ($contact[$name]) $contact[$name] = new_account_id($contact[$name]);
	}
	if (!empty($contact['jpegphoto']))
	{
		$contact['jpegphoto'] = base64_decode($contact['jpegphoto']);
	}
	if (!empty($contact['cat_id']))
	{
		$contact['cat_id'] = new_cat_ids($contact['cat_id']);
	}
	if (!$update_account)
	{
		unset($contact['id']);
		if (preg_match('/^\d+\.vcf$/', $contact['carddav_name'])) unset($contact['carddav_name']);
	}
	unset($contact['.app'], $contact['.id'], $contact['.links'], $contact['etag']);

	return $contacts->save($contact, /*$ignore_acl=*/true, /*$touch_modified=*/false);
}

/*******************************************************************************
 * Calendar support
 ******************************************************************************/

/**
 * Export events
 *
 * @ToDo support $keys parameter to allow to use calendar when following links
 */
function export_calendar(/*$keys*/)
{
	// calendar queries are much slower and returned data is smaller
	// --> use 4 times the global chunk-size (500) speeds up things
	$chunk_size = 4 * $GLOBALS['chunk_size'];

	$cal = new calendar_bo();

	for($chunk=0;
		($entries = $cal->search(array(
			'ignore_acl' => true,
			'filter' => 'everything',
			'daywise' => false,
			'enum_recuring' => false,
			'enum_groups' => false,
			'date_format' => 'Y-m-d H:i:s',
			'offset' => $chunk*$chunk_size,
			'num_rows' => $chunk_size,
			'no_total' => true,
			'order' => 'cal_id ASC',
			'cfs' => array(),	// return all custom fields
		)));
		++$chunk)
	{
		$recurring = array();
		foreach($entries as &$entry)
		{
			unset($entry['tz_id'], $entry['range_start'], $entry['range_end'], $entry['etag'], $entry['recur_date']);

			if (empty($entry['recur_type']))
			{
				unset($entry['recur_enddate'], $entry['recur_exception'], $entry['recur_date']);
			}
			else
			{
				$recurring[] = $entry['id'];
			}
			if (!$entry['alarm']) unset($entry['alarm']);
			if (!$entry['recurrence']) unset($entry['recurrence']);
			if (!$entry['reference']) unset($entry['reference']);
			if (!$entry['special']) unset($entry['special']);
		}
		// query real status exceptions
		if ($recurring)
		{
			foreach($cal->so->db->select($cal->so->user_table, '*', array(
				'cal_id' => $recurring,
				'cal_recur_date != 0',
			), __LINE__, __FILE__, false, '', 'calendar') as $row)
			{
				$uid    = calendar_so::combine_user($row['cal_user_type'], $row['cal_user_id'], $row['cal_user_attendee']);
				$status = calendar_so::combine_status($row['cal_status'],$row['cal_quantity'],$row['cal_role']);
				// really a change to default status, if not skip
				if ($entries[$row['cal_id']]['participants'][$uid] !== $status &&
					// skip deleted markers of uids not in participants
					($row['cal_status'] != 'X' || isset($entries[$row['cal_id']]['participants'][$uid])))
				{
					$date = Api\DateTime::to($row['cal_recur_date'], 'Y-m-d H:i:s');
					$entries[$row['cal_id']]['status_exceptions'][$date][$uid] = $status;
				}
			}
		}
		export_entries($entries, 'calendar', 'id', 'category');
	}
}

function import_calendar($event)
{
	global $exported;

	static $cal = null;
	if (!isset($cal)) $cal = new calendar_boupdate();

	foreach(array('owner', 'creator', 'modifier') as $name)
	{
		if ($event[$name]) $event[$name] = new_account_id($event[$name]);
	}
	if (!empty($event['category']))
	{
		$event['category'] = new_cat_ids($event['category']);
	}
	unset($event['id'], $event['.app'], $event['.id'], $event['.links'], $event['etag']);
	if (preg_match('/^\d+\.vcf$/', $event['caldav_name'])) unset($event['caldav_name']);

	// fix id to series master
	if (!empty($event['reference']))
	{
		$event['reference'] = $exported['calendar'][$event['reference']];
	}

	// fix id of participants in participants
	foreach($event['participants'] as $uid => $status)
	{
		unset($event['participants'][$uid]);

		if (!($new_uid = new_uid($uid)))
		{
			error_log("Participant '$uid' for event #$event[id] not exported --> removed!");
		}
		$event['participants'][$new_uid] = $status;
	}

	$cal_id = $cal->save($event, /*$ignore_acl=*/true, /*$updateTS=*/false);

	// import status exceptions
	if (!empty($event['status_exceptions']))
	{
		foreach($event['status_exceptions'] as $date => $participants)
		{
			foreach($participants as $uid => $status)
			{
				if (($new_uid = new_uid($uid)))
				{
					$cal->set_status($cal_id, $uid, $status, $date,
						/*$ignore_acl=*/true, /*$updateTS=*/false, /*$skip_notification=*/true);
				}
			}
		}
	}

	return $cal_id;
}

/**
 * Get new uid (nummerical id of account, contact or resource)
 *
 * @param string $uid
 * @return string|null
 */
function new_uid($uid)
{
	global $exported;

	switch($uid[0])
	{
		case 'e':	// email --> nothing to do
			$new_uid = $uid;
			break;

		case 'c':	// contacts/addressbook
		case 'r':	// resources
			$app = $uid[0] == 'c' ? 'addressbook' : 'resources';
			if (!isset($exported[$app][substr($uid, 1)]))
			{
				$new_uid = null;
			}
			else
			{
				$new_uid = $uid[0].$exported[$app][substr($uid, 1)];
			}
			break;

		default:
			$new_uid = new_account_id($uid);
			break;
	}
	return $new_uid;
}

/*******************************************************************************
 * InfoLog support
 *
 * Trys to keep ids, as they are displayed and possibly used
 ******************************************************************************/

/**
 * Export InfoLog entries matching given criteria
 *
 * @param array $keys
 */
function export_infolog($keys)
{
	global $chunk_size, $to_export, $exported;

	$infolog_bo = new infolog_bo();

	if (isset($keys[0])) $keys = array('info_id' => $keys);

	$params = array(
		'order' => 'info_id',
		'sort'  => 'ASC',
		'col_filter' => $keys,
		'start' => 0,
		'num_rows' => $chunk_size,
		'custom_fields' => true,
		'filter' => '+deleted',
	);
	for($chunk=0;
		// $params must be passed by reference and if start >= total infolog sets start to 0!
		(!$chunk || ($params['start']=$chunk*$chunk_size) < $params['total'] ? true : false) &&
		($entries = $infolog_bo->search($params, true));	// true: no ACL
		++$chunk)
	{
		foreach($entries as $infolog)
		{
			if (!empty($infolog['info_id_parent']) &&
				!in_array($infolog['info_id_parent'], (array)$to_export['infolog']) &&
				!isset($exported['infolog'][$infolog['info_id_parent']]))
			{
				$to_export['infolog'][] = $infolog['info_id_parent'];
			}
		}
		export_entries($entries, 'infolog', 'info_id', 'info_cat');
	}
}

/**
 * Import Infolog
 *
 * @param array $infolog
 * @return int|false
 */
function import_infolog($infolog)
{
	global $links;

	static $infolog_bo = null;
	if (!isset($infolog_bo)) $infolog_bo = new infolog_bo();

	foreach(array('info_owner', 'info_modifier') as $name)
	{
		if ($infolog[$name]) $infolog[$name] = new_account_id($infolog[$name]);
	}
	foreach($infolog['info_responsible'] as &$responsible)
	{
		$responsible = new_account_id($responsible);
	}
	if (!empty($infolog['info_cat']))
	{
		$infolog['info_cat'] = new_cat_ids($infolog['info_cat']);
	}
	if (!empty($infolog['info_id_parent']))
	{
		$links[$infolog['.app']][$infolog['.id']]['info_id_parent'] = $infolog['info_id_parent'];
	}
	if (!empty($infolog['info_link_id']))
	{
		$links[$infolog['.app']][$infolog['.id']]['info_link_id'] = $infolog['info_link_id'];
	}
	unset($infolog['info_id_parent'], $infolog['info_link_id'],
		$infolog['.app'], $infolog['.id'], $infolog['.links'], $infolog['etag']);

	// check if old id is available on new system, if not remove id
	if ($infolog_bo->read($infolog['info_id'], /*$run_link_id2from=*/false, /*$date_format=*/'ts', /*$ignore_acl=*/true) !== null)
	{
		unset($infolog['info_id']);
		if (preg_match('/^\d+\.ics$/', $infolog['caldav_name'])) unset($infolog['caldav_name']);
	}

	return $infolog_bo->write($infolog, /*$check_defaults=*/false, /*$touch_modified=*/false,
		/*$user2server=*/true, /*$skip_notification=*/true, /*$throw_exception=*/false,
		/*$purge_cfs=*/null, /*$ignore_acl=*/true);
}

/*******************************************************************************
 * Timesheet support
 ******************************************************************************/

/**
 * Export timesheets matching given criteria
 *
 * @param array $keys
 */
function export_timesheet($keys)
{
	global $chunk_size;

	$timesheet = new timesheet_bo();

	if (isset($keys[0])) $keys = array('ts_id' => $keys);
	$keys['ts_owner'] = false;	// no ACL check

	for($chunk=0;
		($entries = $timesheet->search(array(), false, 'ts_id ASC',
			'', '', false, 'AND', array($chunk*$chunk_size, $chunk_size), $keys));
		++$chunk)
	{
		export_entries($entries, 'timesheet', 'ts_id', 'cat_id');
	}
}

/**
 * Import timesheet
 *
 * @param array $timesheet
 * @return int|false
 */
function import_timesheet($timesheet)
{
	static $timesheet_bo = null;
	if (!isset($timesheet_bo)) $timesheet_bo = new timesheet_bo();

	foreach(array('ts_owner', 'ts_modifier') as $name)
	{
		if ($timesheet[$name]) $timesheet[$name] = new_account_id($timesheet[$name]);
	}
	if (!empty($timesheet['cat_id']))
	{
		$timesheet['cat_id'] = new_cat_ids($timesheet['cat_id']);
	}
	unset($timesheet['ts_id'], $timesheet['.app'], $timesheet['.id'], $timesheet['.links']);

	$timesheet_bo->init();
	return !$timesheet_bo->save($timesheet, /*$touch_modified=*/false, /*$ignore_acl=*/true) ?
		$timesheet_bo->data['ts_id'] : false;
}

/*******************************************************************************
 * ACL support
 ******************************************************************************/

/**
 * Export ACL (without group-memberships and anyway not supported apps like SiteMgr)
 */
function export_acl()
{
	global $chunk_size;

	for($chunk=0;
		($rs = $GLOBALS['egw']->db->select('egw_acl', '*', array(
			"acl_appname NOT IN ('phpgw_group','sitemgr')"
		), __LINE__, __FILE__, $chunk*$chunk_size, '', 'api', $chunk_size));
		++$chunk)
	{
		$entries = array();
		foreach($rs as $row)
		{
			$entries[] = $row;
		}
		if (!$entries) break;
		export_entries($entries, 'acl', 'acl_id', 'cat_id');
	}
}

/**
 * Import ACL
 *
 * @param array $entry
 * @return int|false
 */
function import_acl($entry)
{
	if (!($entry['acl_account'] = new_account_id($account_id=$entry['acl_account'], false)))
	{
		error_log("Ignoring ACL for not existing account #$account_id!");
		return false;
	}
	if (is_numeric($entry['acl_location']) &&
		!($entry['acl_account'] = new_account_id($account_id=$entry['acl_account'], false)))
	{
		error_log("Ignoring ACL for not existing account #$account_id!");
		return false;
	}
	// fix cat_id used as location in some apps eg. resources
	if ($entry['acl_location'][0] == 'L' && !($cat_id = new_cat_ids(substr($entry['acl_location'], 1))))
	{
		error_log("Ignoring ACL for not existing category $entry[acl_location]!");
		return false;
	}
	$entry['acl_location'] = 'L'.$cat_id;

	$rights = $entry['acl_rights'];
	unset($entry['.app'], $entry['.id'], $entry['acl_id'], $entry['acl_rights']);

	$GLOBALS['egw']->db->insert('egw_acl', array(
		'acl_rights' => $rights,
	), $entry, __LINE__, __FILE__, 'api');

	// get_last_insert_id() only returns id, if insert does an insert not update
	if (!($acl_id = $GLOBALS['egw']->db->get_last_insert_id('egw_acl', 'acl_id')))
	{
		$acl_id = $GLOBALS['egw']->db->select('egw_acl', 'acl_id', $entry,
			__LINE__, __FILE__, false, '', 'api')->fetchColumn();
	}
	return $acl_id;
}

/*******************************************************************************
 * Resources support
 ******************************************************************************/

/**
 * Export resources matching given criteria
 *
 * @param array $keys
 */
function export_resources($keys)
{
	global $chunk_size;

	// hack to list resource-pictures normally hidden
	Api\Link::$app_register['resources']['find_extra'] = array('hidden' => true);

	$res = new resources_so();
	$res->set_times('string');
	$res->timestamps = array('deleted', 'res_created', 'res_modified');

	if (isset($keys[0])) $keys = array('res_id' => $keys);

	for($chunk=0;
		($entries = $res->search(array(), false, 'res_id ASC',
			'', '', false, 'AND', array($chunk*$chunk_size, $chunk_size), $keys));
		++$chunk)
	{
		$ids = array();
		foreach($entries as $key => &$entry)
		{
			$ids[$entry['res_id']] = $key;

			if ($entry['accessory_of'] <= 0) unset($entry['accessory_of']);
		}
		// read custom fields
		foreach($res->read_customfields(array_keys($ids)) as $id => $cfs)
		{
			$entries[$ids[$id]] += $cfs;
		}
		export_entries($entries, 'resources', 'res_id', 'cat_id');
	}
}

/**
 * Import a resource
 *
 * @param array $resource
 * @return int|false
 */
function import_resources($resource)
{
	global $exported;

	static $res = null;
	if (!isset($res)) $res = new resources_so();

	foreach(array('res_creator', 'res_modifier') as $name)
	{
		if ($resource[$name]) $resource[$name] = new_account_id($resource[$name]);
	}
	if (!empty($resource['cat_id']))
	{
		$resource['cat_id'] = new_cat_ids($resource['cat_id']);
	}
	if (!empty($resource['accessory_of']))
	{
		$resource['accessory_of'] = $exported['resources'][$resource['accessory_of']];
	}
	unset($resource['res_id'], $resource['.app'], $resource['.id'], $resource['.links']);

	return $res->save($resource);
}

/*******************************************************************************
 * User and group homedirectories akt "filemanager" support
 ******************************************************************************/

/**
 * Export user and group homedirectories
 *
 * @param string|array $dirs ='/home'
 */
function export_filemanager($dirs)
{
	if (!$dirs) $dirs = '/home';

	$accounts = new Api\Accounts();

	Api\Vfs::find($dirs, array(
		'depth' => false,	// dirs first
		'need_mime' => true,
	), function($path, $stat) use ($accounts)
	{
		// ignore home directories of no longer existing accounts
		$matches = null;
		if (preg_match('|^/home/([^/]+)|', $path, $matches) && !$accounts->exist($matches[1])) return;

		// not used in EGroupware
		unset($stat['dev'], $stat['nlink'], $stat['rdev'], $stat['atime'], $stat['blksize'], $stat['blocks']);
		$stat['name'] = $stat['path'];
		unset($stat['url'], $stat['path']);

		// export target for symlinks
		if (($stat['mode'] & Api\Vfs\Sqlfs\StreamWrapper::MODE_LINK) == Api\Vfs\Sqlfs\StreamWrapper::MODE_LINK)
		{
			$stat['mime'] = Api\Vfs\Sqlfs\StreamWrapper::SYMLINK_MIME_TYPE;
			$stat['target'] = Api\Vfs::readlink($path);
		}
		// export relative filesystem path for files instead of content
		elseif(($stat['mode'] & Api\Vfs\Sqlfs\StreamWrapper::MODE_FILE) == Api\Vfs\Sqlfs\StreamWrapper::MODE_FILE)
		{
			if (!file_exists($path=Api\Vfs\Sqlfs\StreamWrapper::_fs_path($stat['ino'])))
			{
				error_log("File '$path' NOT found in filesystem under $path --> ignored!");
				return;
			}
			$stat['path'] = substr($path, 1+strlen($GLOBALS['egw_info']['server']['files_dir']));
		}
		$stat['mode'] &= 0777;	// only store real permission bits like in egw_sqlfs table
		export_entries(array($stat), 'filemanager', 'ino');
	});
}

/**
 * Import a resource
 *
 * @param array $stat
 * @return int|false
 */
function import_filemanager($stat)
{
	global $exported_files_dir;

	Api\Vfs::$is_root = true;

	if ($stat['mime'] == Api\Vfs::DIR_MIME_TYPE)
	{
		if (!Api\Vfs::mkdir($stat['name'], $stat['mode'], true))
		{
			return false;
		}
	}
	elseif ($stat['mime'] == Api\Vfs\Sqlfs\StreamWrapper::SYMLINK_MIME_TYPE)
	{
		if (!Api\Vfs::symlink($stat['target'], $stat['name']))
		{
			return false;
		}
	}
	else
	{
		if (!copy($exported_files_dir.'/'.$stat['path'], Api\Vfs::PREFIX.$stat['name']))
		{
			return false;
		}
	}
	Api\Vfs::touch($stat['name'], $stat['mtime']);
	Api\Vfs::chmod($stat['name'], $stat['mode']);
	Api\Vfs::chown($stat['name'], $stat['uid'] ? new_account_id($stat['uid']) : 0);
	Api\Vfs::chgrp($stat['name'], $stat['gid'] ? abs(new_account_id(-$stat['gid'])) : 0);

	return true;	// no need to run an other stat to get fs_id, Vfs handles directories on it's own
}

/*******************************************************************************
 * Records app support
 ******************************************************************************/

/**
 * Export records matching given criteria
 *
 * @param array $keys
 */
function export_records($keys)
{
	global $chunk_size, $exported, $to_export, $records_contact_cols;

	$records_bo = new records_bo();

	if (isset($keys[0])) $keys = array(records_bo::ID_COL => $keys);

	for($chunk=0;
		($entries = $records_bo->search(array(), false, records_bo::ID_COL.' ASC',
			'', '', false, 'AND', array($chunk*$chunk_size, $chunk_size), $keys));
		++$chunk)
	{
		$ids = array();
		foreach($entries as $key => $record)
		{
			foreach($records_contact_cols as $name)
			{
				if (!empty($record[$name]) &&
					!in_array($record[$name], (array)$to_export['addressbook']) &&
					!isset($exported['addressbook'][$record[$name]]))
				{
					$to_export['addressbook'][] = $record[$name];
				}
			}
			$ids[$record[records_bo::ID_COL]] = $key;
		}

		// write customers before records, to have them available on import
		if (($num_contacts = count($to_export['addressbook'])))
		{
			export_addressbook($to_export['addressbook']);
			$to_export['addressbook'] = array_slice($to_export['addressbook'], $num_contacts);
			error_log("- exported $num_contacts entries of addressbook before records");
		}
		// read custom fields
		foreach($records_bo->read_customfields(array_keys($ids)) as $id => $cfs)
		{
			$entries[$ids[$id]] += $cfs;
		}
		export_entries($entries, 'records', records_bo::ID_COL, 'cat_id');
	}
}

/**
 * Import records entry
 *
 * @param array $record
 * @return int
 */
function import_records($record)
{
	global $exported, $records_contact_cols;
	static $records_bo = null;
	if (!isset($records_bo)) $records_bo = new records_bo();

	foreach(array('record_creator', 'record_modifier') as $name)
	{
		if ($record[$name]) $record[$name] = new_account_id($record[$name]);
	}
	if (!empty($record['cat_id']))
	{
		$record['cat_id'] = new_cat_ids($record['cat_id']);
	}
	foreach($records_contact_cols as $name)
	{
		if (!empty($record[$name]))
		{
			if (!isset($exported['addressbook'][$record[$name]]))
			{
				error_log("Importing records:{$record[records_bo::ID_COL]} Customer #{$record[$name]} missing in export!");
			}
			$record[$name] = $exported['addressbook'][$record[$name]];
		}
	}
	unset($record[records_bo::ID_COL], $record[records_bo::ETAG_COL], $record['.app'], $record['.id'], $record['.links']);

	$records_bo->init();
	return !$records_bo->save($record, null, false) ?
		$records_bo->data[records_bo::ID_COL] : false;
}
