<?php
/**
 * EPL: VFS - plugin for addressbook
 *
 * @link http://www.egroupware.org
 * @package epl
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-16 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

namespace EGroupware\Stylite\Vfs\Links;

use EGroupware\Api;
use EGroupware\Api\Acl;

/**
 * VFS plugin for addressbook
 *
 * This plugin creates virtual subdirectories for addressbook to allow an easier navigation:
 *
 * - root contains folder for each addressbooks readable by user
 * - addressbook folder my contains hashes of first letter of n_fileas
 * - matching contacts with their fileas name
 *
 * Both public methods (stat and appdir) work by populating stylite_links_stream_wrapper's stat cache
 * using the following methods:
 *
 * - void StreamWrapper::addStatCache(string $path,array|boolean $data);
 * - void StreamWrapper::addStatCacheDir(string $path,string $mime=self::DIR_MIME_TYPE);
 * - void StreamWrapper::addStatCacheLink(string $path,string $link);
 *
 * In general each virtual directory should have an english name implementing the functionality
 * (gets added with addStatCacheDir) and a translated name added as link to english one (addStatCacheLink).
 * As the links get resolved by the vfs layer, the plugin then has only to deal with the english name.
 */
class Addressbook implements PluginIface
{
	/**
	 * Instance of addressbook_ui
	 *
	 * @var Api\Contacts
	 */
	static protected $ab;

	/**
	 * Application name of this plugin
	 */
	const APP_NAME = 'addressbook';

	/**
	 * Database column / expression used as title
	 */
	const TITLE_COLUMN = 'n_fileas';

	/**
	 * Title of addressbook application (translated)
	 *
	 * @var string
	 */
	static protected $app_title;

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> StreamWrapper::url_stat then looks up the cache
	 *
	 * @param string $id
	 */
	static public function stat($id)
	{
		//echo "<p>".__METHOD__."($id)</p>\n";

		// we just call appdir to preseed the cache with everything from one level higher
		$pattern = StreamWrapper::virtual_parent($id);

		static $appdir_called_for=null;
		if ($pattern !== $appdir_called_for)
		{
			$appdir_called_for = $pattern;
			self::appdir($pattern);
		}
	}

	/**
	 * Return directory entries for addressbook and populate the stat cache
	 *
	 * - root contains folder for each addressbooks readable by user
	 * - addressbook folder my contains hashes of first letter of n_fileas
	 * - matching contacts with their fileas name
	 *
	 * @param string $id
	 * @return array
	 */
	static public function appdir($id)
	{
		//echo '<p>'.__METHOD__."('$id') ".function_backtrace()."</p>\n";
		if (substr($id,-1) == StreamWrapper::VIRTUAL_DELIMITER)
		{
			list($ab,,$rest) = explode(StreamWrapper::VIRTUAL_DELIMITER,$id);
			//echo "<p>".__METHOD__."($id) ab=$ab, letter=$letter</p>\n";
			if (is_null($rest))	// two $ in id
			{
				try
				{
					$ab = self::untranslateAb($ab);
					$letters = self::get_letters($ab);
					if ($letters['all'] <= StreamWrapper::MAX_ENTRIES_NO_HASH)
					{
						return self::entries($id);
					}
					return StreamWrapper::alphabet2cache(self::APP_NAME,$ab.StreamWrapper::VIRTUAL_DELIMITER,null,$letters);
				}
				catch(Api\Exception\WrongParameter $e)
				{
					unset($e);	// not used
					return false;
				}
			}
			else
			{
				return self::entries($id);
			}
		}
		// return virtual directory with addressbooks
		$entries = array();
		$abs = self::$ab->get_addressbooks(Acl::READ,lang('All'));
		foreach($abs as $uid => $label)
		{
			if (substr($uid,-1) == 'p' || !$uid)
			{
				$name = $uid ? 'private' : ($uid === '' ? 'all' : 'accounts');
			}
			elseif ($uid == self::$ab->user && isset($abs[self::$ab->user.'p']))
			{
				$name = 'personal';
			}
			else
			{
				$name = $GLOBALS['egw']->accounts->id2name($uid);
			}
			$link = $name.StreamWrapper::VIRTUAL_DELIMITER;
			StreamWrapper::addStatCacheDir(self::APP_NAME.'/'.$link,'egw/'.self::APP_NAME);
			if ($name != $label)
			{
				StreamWrapper::addStatCacheLink(self::APP_NAME.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,$link);
			}
			if (self::$app_title != self::APP_NAME)
			{
				StreamWrapper::addStatCacheLink(self::$app_title.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,'../'.self::APP_NAME.'/'.$link);

				if ($name != $label) StreamWrapper::addStatCacheLink(self::$app_title.'/'.$link,'../'.self::APP_NAME.'/'.$link);
			}
			$entries[] = $label.StreamWrapper::VIRTUAL_DELIMITER;
		}
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$id') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Translate addessbook names back to their english (lowercase) counterparts
	 *
	 * @param string $_ab
	 * @return string
	 */
	static private function untranslateAb($_ab)
	{
		static $abs = array('accounts','private','personal','all');

		$ab = strtolower($_ab);

		if (!in_array($ab,$abs))
		{
			$strlen_group = strlen($group = lang('group'));
			if ($GLOBALS['egw']->accounts->name2id($_ab))
			{
				$ab = $_ab;
			}
			elseif(trim(substr($_ab,0,1+$strlen_group)) == trim($group).' ' && $GLOBALS['egw']->accounts->name2id(trim(substr($_ab,1+$strlen_group)),'account_lid','g'))
			{
				$ab = substr($_ab,0,$strlen_group);
			}
			else
			{
				foreach($abs as $name)
				{
					if (strtolower(lang($name)) == $ab)
					{
						$ab = $name;
						break;
					}
				}
				if (!in_array($ab,$abs))
				{
					throw new Api\Exception\WrongParameter(__METHOD__."('$ab') unknown Addressbook!");
				}
			}
		}
		return $ab;
	}

	/**
	 * Create an owner filter from addressbook name
	 *
	 * @param unknown_type $ab
	 * @return unknown
	 */
	static private function ab2filter($ab)
	{
		switch($ab)
		{
			case 'all':
				$filter = null;
				break;
			case 'accounts':
				$filter = array('owner' => '0');
				break;
			case 'personal':
			case 'private':
				$filter = array('owner' => self::$ab->user,'private' => (int)($ab == 'private'));
				break;
			default:
				if (!($uid = $GLOBALS['egw']->accounts->name2id($ab)))
				{
					error_log(__METHOD__."('$ab') unknown addressbook!");
					return false;
				}
				$filter = array('owner' => $uid);
				break;
		}
		// do not show accounts, if disabled in user prefs
		if ($GLOBALS['egw_info']['user']['preferences']['addressbook']['hide_accounts'] === '1')
		{
			$filter['account_id'] = null;
		}
		return $filter;
	}

	/**
	 * Return $app's entries and add them to the cache
	 *
	 * @param string $pattern
	 * @return array|boolean false on error (eg. no app rights) or array with file-names
	 */
	static private function entries($pattern)
	{
		list($ab,$letter,$rest) = explode(StreamWrapper::VIRTUAL_DELIMITER,$pattern);

		$filter = self::ab2filter($ab);

		// char-filter
		if (!is_null($rest) && $letter != 'all' && $letter != lang('all'))
		{
			if ($letter === '')
			{
				$filter[] = '('.self::TITLE_COLUMN.' IS NULL OR '.self::TITLE_COLUMN."='')";
			}
			else
			{
				$filter[] = self::TITLE_COLUMN.' LIKE '.$GLOBALS['egw']->db->quote($letter.'%');
			}
		}
		//echo __METHOD__."($pattern) filter="; _debug_array($filter);
		$entries = array();
		foreach((array)self::$ab->search('',array('id','n_prefix','n_given','n_middle','n_family','n_suffix','n_fn','org_name','org_unit','adr_one_locality',self::TITLE_COLUMN),self::TITLE_COLUMN,'','',false,'AND',false,$filter) as $row)
		{
			if ($row === false) continue;
			$id = $row['id'];
			if (empty($row[self::TITLE_COLUMN]))
			{
				$row[self::TITLE_COLUMN] = self::$ab->fileas($row);
			}
			$entries[] = $name = StreamWrapper::entry2name(self::APP_NAME,$id,$row[self::TITLE_COLUMN]);
			StreamWrapper::entry2cache(self::APP_NAME,$id,$name,null,$pattern);
		}
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$pattern') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Get the first letters used in title-column of addressbook $ab
	 *
	 * @param string $ab addressbook name
	 * @return array with letter => number of entries pairs (NULL and '' are returned as '')
	 */
	static private function get_letters($ab)
	{
		static $letters=null,$last_ab=null;

		if ($ab != $last_ab)
		{
			$last_ab = $ab;

			$letters = array();
			$total = 0;
			$filter = self::ab2filter($ab);
			foreach((array)self::$ab->search(array(),array('SUBSTR(UPPER('.self::TITLE_COLUMN.'),1,1) AS letter','COUNT(*) AS num'),'GROUP BY letter ORDER BY letter','','',false,'AND',false,$filter) as $row)
			{
				if ($row === false) continue;
				$letters[(string)$row['letter']] += $row['num'];	// (string) puts both NULL and '' under ''
				$total += $row['num'];
			}
			$letters['all'] = $total;
			//echo __METHOD__."($ab) ".function_backtrace(); _debug_array($letters);
		}
		return $letters;
	}

	/**
	 * Init our static properties
	 *
	 */
	static function init_static()
	{
		self::$ab = new Api\Contacts();

		// add addressbook translations
		Api\Translation::add_app(self::APP_NAME);

		self::$app_title = $GLOBALS['egw_info']['user']['apps'][self::APP_NAME]['title'];
	}
}

Addressbook::init_static();
