<?php
/**
 * EPL: VFS - plugin for calendar
 *
 * @link http://www.egroupware.org
 * @package epl
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-16 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

namespace EGroupware\Stylite\Vfs\Links;

use EGroupware\Api;

// explicitly import old classes
use calendar_bo;

/**
 * VFS plugin for calendar
 *
 * This plugin creates virtual subdirectories for calendar to allow an easier navigation:
 *
 * - root contains folder for each calendar readable by the user
 * - calendar folder contain hashes (year or year-month)
 *
 * Both public methods (stat and appdir) work by populating stylite_links_stream_wrapper's stat cache
 * using the following methods:
 *
 * - void StreamWrapper::addStatCache(string $path,array|boolean $data);
 * - void StreamWrapper::addStatCacheDir(string $path,string $mime=self::DIR_MIME_TYPE);
 * - void StreamWrapper::addStatCacheLink(string $path,string $link);
 *
 * In general each virtual directory should have an english name implementing the functionality
 * (gets added with addStatCacheDir) and a translated name added as link to english one (addStatCacheLink).
 * As the links get resolved by the vfs layer, the plugin then has only to deal with the english name.
 *
 * @todo Implement structure with 3 levels: owner/$year$/month$ as year-month hash takes some time,
 *       year and month of one year might be faster
 */
class Calendar implements PluginIface
{
	/**
	 * Use only year as hash, eg. 2009
	 */
	const HASH_YEAR = '%Y';
	/**
	 * Use year and month as hash, eg. 2009-05
	 */
	const HASH_YEAR_MONTH = '%Y-%m';
	/**
	 * Should we hash by year or year and month, use HASH_YEAR or HASH_YEAR_MONTH
	 *
	 * @var string
	 */
	static $hash_format = self::HASH_YEAR;

	/**
	 * Column used for the hash
	 *
	 * @var string
	 */
	static $hash_column = 'cal_start';

	/**
	 * Instance of calendar
	 *
	 * @var calendar_bo
	 */
	static protected $cal;

	/**
	 * Application name of this plugin
	 */
	const APP_NAME = 'calendar';

	/**
	 * Title of addressbook application (translated)
	 *
	 * @var string
	 */
	static protected $app_title;

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> StreamWrapper::url_stat then looks up the cache
	 *
	 * @param string $id
	 */
	static public function stat($id)
	{
		//echo "<p>".__METHOD__."($id)</p>\n";

		// we just call appdir to preseed the cache with everything from one level higher
		$pattern = StreamWrapper::virtual_parent($id);

		static $appdir_called_for=null;
		if ($pattern !== $appdir_called_for)
		{
			$appdir_called_for = $pattern;
			self::appdir($pattern);
		}
	}

	/**
	 * Return directory entries for addressbook and populate the stat cache
	 *
	 * - root contains folder for each calendar readable by the user
	 * - calendar folder contain hashes (year or year-month)
	 *
	 * @param string $id
	 * @return array
	 */
	static public function appdir($id)
	{
		if ($id)
		{
			list($owner,$hash) = explode(StreamWrapper::VIRTUAL_DELIMITER,$id);

			if (!($account_id = self::owner2id($owner)))
			{
				return null;
			}
		}
		//echo '<p>'.__METHOD__."('$id') --> owner=$owner, hash=$hash</p>\n";
		if ($owner && $hash)	// --> add events
		{
			$entries = self::entries($owner,$hash,$account_id);
		}
		elseif($owner)	// --> add hashes
		{
			$hashes = self::get_hashes($account_id);

			$entries = StreamWrapper::hashes2cache($hashes,self::APP_NAME,
				$owner.StreamWrapper::VIRTUAL_DELIMITER);
		}
		else	// --> add calendars
		{
			foreach(self::$cal->list_cals() as $label => $data)
			{
				if (!($name = $GLOBALS['egw']->accounts->id2name($data['grantor']))) continue;
				$link = $name.StreamWrapper::VIRTUAL_DELIMITER;
				StreamWrapper::addStatCacheDir(self::APP_NAME.'/'.$link,'egw/'.self::APP_NAME);
				if ($name != $label)
				{
					StreamWrapper::addStatCacheLink(self::APP_NAME.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,$link);
				}
				if (self::$app_title != self::APP_NAME)
				{
					StreamWrapper::addStatCacheLink(self::$app_title.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,'../'.self::APP_NAME.'/'.$link);

					if ($name != $label) StreamWrapper::addStatCacheLink(self::$app_title.'/'.$link,'../'.self::APP_NAME.'/'.$link);
				}
				$entries[] = $label.StreamWrapper::VIRTUAL_DELIMITER;
			}
		}
		//_debug_array($entries);
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$id') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Convert an calendar owner into the numeric account_id
	 *
	 * @param int|string $owner account_id, account_lid or result from Api\Accounts::username($id)
	 * @return int account_id or NULL if unknown $owner or no rights
	 */
	static private function owner2id($owner)
	{
		if (is_numeric($owner))
		{
			$id = (int)$owner;
		}
		else
		{
			$id = $GLOBALS['egw']->accounts->name2id($owner);
		}
		if ($id && isset(self::$cal->grants[$id]))
		{
			$ret = $id;
		}
		else
		{
			foreach(array_keys(self::$cal->grants) as $id)
			{
				if ($owner == Api\Accounts::username($id))
				{
					$ret = $id;
					break;
				}
			}
		}
		//echo "<p>".__METHOD__."($owner) = ".array2string($ret)."</p>\n";
		return $ret;
	}

	/**
	 * Get the main projects with a given state and evtl. year
	 *
	 * @param string $owner calendar owner
	 * @param string $hash year or year-month
	 * @param int $account_id calendar owner
	 * @return array
	 */
	static private function entries($owner,$hash,$account_id)
	{
		//echo "<p>".__METHOD__."($account_id,$hash)</p>\n";
		$entries = array();

		if ($hash == 'all' || $hash == lang('all'))
		{
			$start = $end = 0;
		}
		else
		{
			list($year,$month_start) = explode('-',$hash);
			if (!$month_start)
			{
				$month_start = 1;
				$month_end   = 12;
			}
			else
			{
				$month_end = $month_start;
			}
			$start = mktime(0,0,0,$month_start,1,$year);
			$end   = mktime(0,0,0,$month_end+1,1,$year);
		}
		foreach(self::$cal->search(array(
			'users' => $account_id,
			'start' => $start,
			'end'   => $end,
			'enum_recuring' => true,
			'enum_groups' => true,
			'order' => 'cal_start ASC',
			)) as $event)
		{
			$id = $event['id'];
			//$title = self::$cal->link_title($event);
			$title = date('Y-m-d H.i',$event['start']).' '.$event['title'];
			$entries[] = $name = StreamWrapper::entry2name(self::APP_NAME,$id,$title);
			StreamWrapper::entry2cache(self::APP_NAME,$id,$name,null,$owner.$hash);
		}
		return $entries;
	}

	/**
	 * Get the hash values (eg. creation years) of existing projects of the given status
	 *
	 * @param int $account_id
	 * @return array
	 */
	static private function get_hashes($account_id)
	{
		static $cache=array();
		$hashes =& $cache[$account_id.':'.self::$hash_format];

		if(!isset($hashes))
		{
			//echo "<p>".__METHOD__."($owner) hash_format='".self::$hash_format."' ".function_backtrace()."</p>\n";
			$hashes = array();
			foreach(self::$cal->search(array(
				'users' => $account_id,
				'start' => 0,
				'end'   => 0,
				'order' => 'hash DESC',
				'cols'  => array(
					$GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime(self::$hash_column),self::$hash_format).' AS hash',
					'COUNT(*) AS num',
				),
				'append' => 'GROUP BY hash',
			)) as $row)
			{
				// if more then MAX_ENTRIES_ALL=200 entries in one hash dir, automatic seitch to YEAR-MONTH hashes
				if (self::$hash_format == self::HASH_YEAR && $row['num'] > StreamWrapper::MAX_ENTRIES_ALL)
				{
					self::$hash_format = self::HASH_YEAR_MONTH;
					return self::get_hashes($account_id);
				}
				$hashes[$row['hash']] = $row['num'];
				$total += $row['num'];
			}
			$hashes['all'] = (int)$total;
			//_debug_array($hashes);
		}
		return $hashes;
	}

	/**
	 * Init our static properties
	 *
	 */
	static function init_static()
	{
		self::$cal = new calendar_bo();

		// add addressbook translations
		Api\Translation::add_app(self::APP_NAME);

		self::$app_title = $GLOBALS['egw_info']['user']['apps'][self::APP_NAME]['title'];
	}
}

Calendar::init_static();
