<?php
/**
 * EPL: VFS - plugin for projectmanager
 *
 * @link http://www.egroupware.org
 * @package epl
 * @subpackage vfs
 * @author Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @copyright (c) 2009-16 by Ralf Becker <RalfBecker-AT-outdoor-training.de>
 * @version $Id$
 */

namespace EGroupware\Stylite\Vfs\Links;

use EGroupware\Api;

// explicitly import old classes
use projectmanager_bo;

/**
 * VFS plugin for projectmanager
 *
 * This plugin creates virtual subdirectories for projectmanager to allow an easier navigation:
 *
 * - root contains all active main projects (or hashes for them) plus archive$, nonactive$ and template$ folder
 * - archive$, nonactive$ and template$ contains subfolders for each year (or year-month) we have archived projects
 * - project folder allways contain links to subprojects (as these are regular links anyway)
 *
 * Both public methods (stat and appdir) work by populating stylite_links_stream_wrapper's stat cache
 * using the following methods:
 *
 * - void StreamWrapper::addStatCache(string $path,array|boolean $data);
 * - void StreamWrapper::addStatCacheDir(string $path,string $mime=self::DIR_MIME_TYPE);
 * - void StreamWrapper::addStatCacheLink(string $path,string $link);
 *
 * In general each virtual directory should have an english name implementing the functionality
 * (gets added with addStatCacheDir) and a translated name added as link to english one (addStatCacheLink).
 * As the links get resolved by the vfs layer, the plugin then has only to deal with the english name.
 */
class Projectmanager implements PluginIface
{
	/**
	 * Use only year as hash, eg. 2008
	 */
	const HASH_YEAR = '%Y';
	/**
	 * Use year and month as hash, eg. 2008-05
	 */
	const HASH_YEAR_MONTH = '%Y-%m';
	/**
	 * Should we hash by year or year and month, use HASH_YEAR or HASH_YEAR_MONTH
	 *
	 * @var string
	 */
	static $hash_format = self::HASH_YEAR;

	/**
	 * Column used for the hash
	 *
	 * @var string
	 */
	static $hash_column = 'pm_created';	// 'pm_modified', 'pm_planned_start', ...

	/**
	 * Instance of projectmanager
	 *
	 * @var projectmanager_bo
	 */
	static protected $pm;

	/**
	 * Application name of this plugin
	 */
	const APP_NAME = 'projectmanager';

	/**
	 * Title of addressbook application (translated)
	 *
	 * @var string
	 */
	static protected $app_title;

	/**
	 * Status of PM (should come from PM, but is only in UI)
	 *
	 * @var array
	 */
	static protected $status = array('active','nonactive','archive','template');

	/**
	 * Stat something in the application directory
	 *
	 * An implementation can eg. call the appdir method of one level higher, to populate the stat cache
	 * with everything allowed on that level --> StreamWrapper::url_stat then looks up the cache
	 *
	 * @param string $id
	 * @param array $params =array() current only key "backup"
	 */
	static public function stat($id, $params=array())
	{
		//echo "<p>".__METHOD__."($id)</p>\n";

		// we just call appdir to preseed the cache with everything from one level higher
		$pattern = StreamWrapper::virtual_parent($id);

		static $appdir_called_for=null;
		if ($pattern !== $appdir_called_for)
		{
			$appdir_called_for = $pattern;
			self::appdir($pattern, $params);
		}
	}

	/**
	 * Return directory entries for projectmanager and populate the stat cache
	 *
	 * - root contains all active main projects (or hashes for them) plus archive$, nonactive$ and template$ folder
	 * - archive$, nonactive$ and template$ contains subfolders for each year (or year-month) we have archived projects
	 * - project folder allways contain links to subprojects (as these are regular links anyway)
	 *
	 * @param string $id
	 * @param array $params =array() current only key "backup"
	 * @return array
	 */
	static public function appdir($id, $params=array())
	{
		$status = !empty($params['backup']) ? 'backup' : 'active';
		if ($id)
		{
			list($year,$hash) = explode(StreamWrapper::VIRTUAL_DELIMITER,$id);

			if ($year == lang('all')) $year = 'all';
			else if (!(int)$year)
			{
				$status = $year;
			}
			if ((int)$year || $year == 'all')	// year folder for active
			{
				$hash = $year;
				if ($status != 'backup') $status = 'active';
			}
		}
		// force year hashes for backup
		if (!$hash) $hashes = self::get_hashes($status, $hash, !empty($params['backup']));

		if ($hash || isset($hashes['all']) && $hashes['all'] <= StreamWrapper::MAX_ENTRIES_NO_HASH)
		{
			$entries = self::entries($status,$hash);
		}
		else
		{
			$entries = StreamWrapper::hashes2cache($hashes,self::APP_NAME,
				!in_array($status, array('backup', 'active')) ? $status.StreamWrapper::VIRTUAL_DELIMITER : '');
		}
		if ($status == 'active' && !$hash)	// add other states
		{
			foreach(array('nonactive','archive','template') as $name)
			{
				$label = /*'zzz '.*/lang($name);
				$link = $name.StreamWrapper::VIRTUAL_DELIMITER;
				StreamWrapper::addStatCacheDir(self::APP_NAME.'/'.$link,'egw/'.self::APP_NAME);
				if ($name != $label)
				{
					StreamWrapper::addStatCacheLink(self::APP_NAME.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,$link);
				}
				if (self::$app_title != self::APP_NAME)
				{
					StreamWrapper::addStatCacheLink(self::$app_title.'/'.$label.StreamWrapper::VIRTUAL_DELIMITER,'../'.self::APP_NAME.'/'.$link);

					if ($name != $label) StreamWrapper::addStatCacheLink(self::$app_title.'/'.$link,'../'.self::APP_NAME.'/'.$link);
				}
				$entries[] = $label.StreamWrapper::VIRTUAL_DELIMITER;
			}
		}
		if (StreamWrapper::DEBUG > 1) error_log(__METHOD__."('$id') = ".array2string($entries));
		return $entries;
	}

	/**
	 * Get the main projects with a given state and evtl. year
	 *
	 * @param string $status ='active' active, inactive, archve or template
	 * @param string $hash =null creation year (or year-month) of project
	 * @return array
	 */
	static private function entries($status='active',$hash=null)
	{
		//echo "<p>".__METHOD__."($status,$hash)</p>\n";
		$entries = array();
		foreach((array)self::$pm->search('',self::$pm->table_name.'.pm_id AS pm_id,pm_number,pm_title','pm_number DESC',
			'','',false,'AND',false,self::get_filter($status,$hash)) as $project)
		{
			$id = $project['pm_id'];
			$entries[] = $name = StreamWrapper::entry2name(self::APP_NAME,$id,self::$pm->link_title($project));
			StreamWrapper::entry2cache(self::APP_NAME,$id,$name,null,$hash);
		}
		return $entries;
	}

	/**
	 * Get the hash values (eg. creation years) of existing projects of the given status
	 *
	 * @param string $status ='active' active, inactive, archve or template
	 * @param string $hash =''
	 * @param boolean $force_year_hash =false
	 * @return array
	 */
	static private function get_hashes($status='active', $hash='', $force_year_hash=false)
	{
		//echo "<p>".__METHOD__."($status)</p>\n";

		static $cache = array();
		$hashes =& $cache[$status.$hash.':'.self::$hash_format];

		if(!isset($hashes))
		{
			$hashes = array();
			foreach((array)self::$pm->search('',array(
				$GLOBALS['egw']->db->date_format($GLOBALS['egw']->db->from_unixtime(self::$hash_column),self::$hash_format).' AS hash',
				'COUNT(*) AS num',
			),'GROUP BY hash ORDER BY hash DESC','','',false,'AND',false,self::get_filter($status,$hash)) as $row)
			{
				// if more then MAX_ENTRIES_ALL=200 entries in one hash dir, automatic switch to YEAR-MONTH hashes
				if (self::$hash_format == self::HASH_YEAR && $row['num'] > StreamWrapper::MAX_ENTRIES_ALL && !$force_year_hash)
				{
					self::$hash_format = self::HASH_YEAR_MONTH;
					return self::get_hashes($status);
				}
				$hashes[$row['hash']] = $row['num'];
				$total += $row['num'];
			}
			if ($status != 'backup') $hashes['all'] = (int)$total;
		}
		//_debug_array($hashes);
		return $hashes;
	}

	/**
	 * Get PM search filter from given $status and $hash
	 *
	 * @param string $status ='active'
	 * @param int $hash =null
	 * @return array
	 */
	static private function get_filter($status='active',$hash=null)
	{
		if (!in_array($status,self::$status))
		{
			foreach(self::$status as $s)
			{
				if ($status == lang($s))
				{
					$status = $s;
					break;
				}
			}
		}
		$filter = $status == 'backup' ? array() : array(
			'pm_status' => $status,
			'subs_or_mains' => 'mains',
		);
		if ((int)$hash)
		{
			list($year,$month_start) = explode('-',$hash);
			if (!$month_start)
			{
				$month_start = 1;
				$month_end   = 12;
			}
			else
			{
				$month_end = $month_start;
			}
			$filter[] = mktime(0,0,0,$month_start,1,$year).' <= '.self::$hash_column;
			$filter[] = self::$hash_column.' < '.mktime(0,0,0,$month_end+1,1,$year);
		}
		return $filter;
	}

	/**
	 * Init our static properties
	 *
	 */
	static function init_static()
	{
		self::$pm = new projectmanager_bo();

		// add addressbook translations
		Api\Translation::add_app(self::APP_NAME);

		self::$app_title = $GLOBALS['egw_info']['user']['apps'][self::APP_NAME]['title'];
	}
}

Projectmanager::init_static();
